# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import re
import requests
import random
import string

from knack.log import get_logger
from knack.prompting import (
    prompt,
    prompt_pass
)
from azure.mgmt.core.tools import (
    parse_resource_id,
)
from azure.cli.core import telemetry
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.core.azclierror import (
    ValidationError,
    InvalidArgumentValueError,
    RequiredArgumentMissingError,
)

from ._utils import (
    run_cli_cmd,
    get_object_id_of_current_user,
    is_valid_resource_id
)
from ._resource_config import (
    AUTH_TYPE,
    CLIENT_TYPE,
    RESOURCE,
    SOURCE_RESOURCES,
    TARGET_RESOURCES,
    SOURCE_RESOURCES_PARAMS,
    SOURCE_RESOURCES_OPTIONAL_PARAMS,
    SOURCE_RESOURCES_CREATE_PARAMS,
    TARGET_RESOURCES_PARAMS,
    AUTH_TYPE_PARAMS,
    SUPPORTED_AUTH_TYPE,
    LOCAL_CONNECTION_RESOURCE,
    LOCAL_CONNECTION_PARAMS,
    SPRING_APP_DEPLOYMENT_RESOURCE,
    WEB_APP_SLOT_RESOURCE,
    OPT_OUT_OPTION,
)


logger = get_logger(__name__)


def get_source_resource_name(cmd):
    '''Get source resource name
    e.g, az webapp connection list: => RESOURCE.WebApp
    '''
    source = None
    source_name = cmd.name.split(' ')[0]
    if source_name == RESOURCE.Local.value.lower():
        return RESOURCE.Local
    for item in SOURCE_RESOURCES:
        if item.value.lower() == source_name.lower():
            source = item
    return source


def get_target_resource_name(cmd):
    '''Get target resource name
    e.g, az webapp connection create postgres: => RESOURCE.Postgres
    '''
    target = None
    target_name = cmd.name.split(' ')[-1]
    for item in TARGET_RESOURCES:
        if item.value.lower() == target_name.lower():
            target = item
    return target


def get_resource_type_by_id(resource_id):
    '''Get source or target resource type by resource id
    '''
    target_type = None
    all_resources = {}
    all_resources.update(SOURCE_RESOURCES)
    all_resources.update(TARGET_RESOURCES)
    for _type, _id in all_resources.items():
        matched = re.match(get_resource_regex(_id), resource_id, re.IGNORECASE)
        if matched:
            target_type = _type
    return target_type


def get_resource_regex(resource):
    '''Replace '{...}' with '[^/]*' for regex matching
    '''
    regex = resource
    for item in re.findall(r'(\{[^\{\}]*\})', resource):
        regex = regex.replace(item, '[^/]*')
    return regex


def check_required_args(resource, cmd_arg_values):
    '''Check whether a resource's required arguments are in cmd_arg_values
    '''
    args = re.findall(r'\{([^\{\}]*)\}', resource)

    if 'subscription' in args:
        args.remove('subscription')
    for arg in args:
        if not cmd_arg_values.get(arg, None):
            return False
    return True


def get_fabric_access_token():
    get_fabric_token_cmd = 'az account get-access-token --output json --resource https://api.fabric.microsoft.com/'
    return run_cli_cmd(get_fabric_token_cmd).get('accessToken')


def generate_fabric_connstr_props(target_id):
    fabric_token = get_fabric_access_token()
    headers = {"Authorization": "Bearer {}".format(fabric_token)}
    response = requests.get(target_id, headers=headers)

    if response:
        response_json = response.json()

        if "properties" in response_json:
            properties = response_json["properties"]
            if "databaseName" in properties and "serverFqdn" in properties:
                return {
                    "Server": properties["serverFqdn"],
                    "Database": properties["databaseName"]
                }

    return None


def generate_connection_name(cmd):
    '''Generate connection name for users if not provided
    '''
    target = get_target_resource_name(cmd).value.replace('-', '')
    randstr = ''.join(random.sample(string.ascii_lowercase + string.digits, 5))
    name = '{}_{}'.format(target, randstr)

    logger.warning('Connection name is not specified, use generated one: --connection %s', name)
    return name


def get_client_type(cmd, namespace):
    '''Infer client type from source resource
    '''

    def _infer_webapp(source_id):
        value_type_map = {
            'python': CLIENT_TYPE.Python,
            'dotnet': CLIENT_TYPE.Dotnet,
            'dotnetcore': CLIENT_TYPE.Dotnet,
            'php': CLIENT_TYPE.Php,
            'java': CLIENT_TYPE.Java,
            'nodejs': CLIENT_TYPE.Nodejs,
            'node': CLIENT_TYPE.Nodejs,
            'ruby': CLIENT_TYPE.Ruby,
        }

        prop_type_map = {
            'javaVersion': CLIENT_TYPE.Java,
            'netFrameworkVersion': CLIENT_TYPE.Dotnet,
            'pythonVersion': CLIENT_TYPE.Python,
            'nodeVersion': CLIENT_TYPE.Nodejs,
            'phpVersion': CLIENT_TYPE.Php,
        }

        client_type = None
        try:
            output = run_cli_cmd('az webapp show --id "{}" -o json'.format(source_id))
            prop = output.get('siteConfig').get('linuxFxVersion', None) or\
                output.get('siteConfig').get('windowsFxVersion', None)
            # use 'linuxFxVersion' and 'windowsFxVersion' property to decide
            if prop:
                language = prop.split('|')[0].lower()
                client_type = value_type_map.get(language, None)
            # use '*Version' property to decide
            if client_type is None:
                for prop, _type in prop_type_map.items():
                    if output.get('siteConfig', {}).get(prop, None) is not None:
                        client_type = _type
                        break
        except Exception:  # pylint: disable=broad-except
            pass

        return client_type

    def _infer_springcloud(source_id):
        client_type = CLIENT_TYPE.SpringBoot
        try:
            segments = parse_resource_id(source_id)
            output = run_cli_cmd('az spring app show -g "{}" -s "{}" -n "{}"'
                                 ' -o json'.format(segments.get('resource_group'), segments.get('name'),
                                                   segments.get('child_name_1')))
            prop_val = output.get('properties')\
                             .get('activeDeployment')\
                             .get('properties')\
                             .get('deploymentSettings')\
                             .get('runtimeVersion')

            language = prop_val.split('_')[0].lower()
            if language == 'java':
                client_type = CLIENT_TYPE.Java
            elif language == 'netcore':
                client_type = CLIENT_TYPE.Dotnet
        except Exception:   # pylint: disable=broad-except
            pass

        return client_type

    # fallback to use None as client type
    client_type = None
    if 'webapp' in cmd.name:
        client_type = _infer_webapp(namespace.source_id)
    elif 'spring-cloud' in cmd.name or 'spring' in cmd.name:
        client_type = _infer_springcloud(namespace.source_id)

    method = 'detected'
    if client_type is None:
        client_type = CLIENT_TYPE.Blank
        method = 'default'

    logger.warning('Client type is not specified, use %s one: --client-type %s', method, client_type.value)
    return client_type.value


def interactive_input(arg, hint, cmd):
    '''Get interactive inputs from users
    '''
    value = None
    cmd_value = None
    if arg == 'secret_auth_info':
        if 'mongodb-atlas' in cmd.name:
            secret = prompt_pass('Connection string of cluster (--secret secret=): ')
            value = {
                'name': 'NA',
                'secret_info': {
                    'secret_type': 'rawValue',
                    'value': secret
                },
                'auth_type': 'secret'
            }
            cmd_value = 'secret={}'.format('*' * len(secret))
        else:
            name = prompt('User name of database (--secret name=): ')
            secret = prompt_pass('Password of database (--secret secret=): ')
            value = {
                'name': name,
                'secret_info': {
                    'secret_type': 'rawValue',
                    'value': secret
                },
                'auth_type': 'secret'
            }
            cmd_value = 'name={} secret={}'.format(name, '*' * len(secret))
    elif arg == 'service_principal_auth_info_secret':
        client_id = prompt('ServicePrincipal client-id (--service-principal client_id=): ')
        object_id = prompt('Enterprise Application object-id (--service-principal object-id=): ')
        secret = prompt_pass('ServicePrincipal secret (--service-principal secret=): ')
        value = {
            'client_id': client_id,
            'object-id': object_id,
            'secret': secret,
            'auth_type': 'servicePrincipalSecret'
        }
        cmd_value = 'client-id={} principal-id={} secret={}'.format(client_id, object_id, '*' * len(secret))
    elif arg == 'user_identity_auth_info':
        client_id = prompt('UserAssignedIdentity client-id (--user-identity client_id=): ')
        subscription_id = prompt('UserAssignedIdentity subscription-id (--user-identity subs_id=): ')
        value = {
            'client_id': client_id,
            'subscription_id': subscription_id,
            'auth_type': 'userAssignedIdentity'
        }
        cmd_value = 'client-id={} subscription-id={}'.format(client_id, subscription_id)
    elif arg == 'user_account_auth_info':
        object_id = prompt(
            'User Account object-id (--user-account object-id=): ')
        value = {
            'auth_type': 'userAccount',
            'principal_id': object_id
        }
        cmd_value = 'object-id={}'.format(object_id)
    else:
        value = prompt('{}: '.format(hint))
        cmd_value = value

    # check blank value
    if isinstance(value, dict):
        for sub_val in value.values():
            if not sub_val:
                raise RequiredArgumentMissingError('{} should not be blank'.format(hint))
    elif not value:
        raise RequiredArgumentMissingError('{} should not be blank'.format(hint))

    return value, cmd_value


def get_local_context_value(cmd, arg):
    '''Get local context values
    '''
    groups = ['all', 'cupertino', 'serviceconnector', 'postgres']
    arg_map = {
        'source_resource_group': ['resource_group_name'],
        'target_resource_group': ['resource_group_name'],
        'server': ['server_name', "server"],
        'database': ['database_name', "database"],
        'site': ['webapp_name']
    }
    for group in groups:
        possible_args = arg_map.get(arg, [arg])
        for item in possible_args:
            if cmd.cli_ctx.local_context.get(group, item):
                return cmd.cli_ctx.local_context.get(group, item)
    return None


def opt_out_auth(namespace):
    '''Validate if config and auth are both opted out
    '''
    opt_out_list = getattr(namespace, 'opt_out_list', None)
    if opt_out_list is not None and \
            OPT_OUT_OPTION.AUTHENTICATION.value in opt_out_list:
        return True
    return False


def intelligent_experience(cmd, namespace, missing_args):
    '''Use local context and interactive inputs to get arg values
    '''
    cmd_arg_values = {}
    # use commandline source/target resource args
    for arg in missing_args:
        if getattr(namespace, arg, None) is not None:
            cmd_arg_values[arg] = getattr(namespace, arg)

    # for auth info without additional parameters
    if 'secret_auth_info_auto' in missing_args:
        cmd_arg_values['secret_auth_info_auto'] = {
            'auth_type': 'secret'
        }
        logger.warning('Auth info is not specified, use default one: --secret')
    elif 'system_identity_auth_info' in missing_args:
        cmd_arg_values['system_identity_auth_info'] = {
            'auth_type': 'systemAssignedIdentity'
        }
        logger.warning('Auth info is not specified, use default one: --system-identity')
        if opt_out_auth(namespace):
            logger.warning('Auth info is only used to generate configurations. %s',
                           'Skip enabling identity and role assignments.')
    elif 'user_account_auth_info' in missing_args:
        cmd_arg_values['user_account_auth_info'] = {
            'auth_type': 'userAccount'
        }
        logger.warning(
            'Auth info is not specified, use default one: --user-account')
    if cmd.cli_ctx.local_context.is_on:
        # arguments found in local context
        context_arg_values = {}
        for arg in missing_args:
            if arg not in cmd_arg_values:
                if get_local_context_value(cmd, arg):
                    context_arg_values[arg] = get_local_context_value(cmd, arg)

        # apply local context arguments
        param_str = ''
        for k, v in context_arg_values.items():
            option = missing_args[k].get('options')[0]
            value = v
            param_str += '{} {} '.format(option, value)
        if param_str:
            logger.warning('Apply local context arguments: %s', param_str.strip())
            cmd_arg_values.update(context_arg_values)

    # arguments from interactive inputs
    param_str = ''
    for arg in missing_args:
        if arg not in cmd_arg_values:
            hint = '{} ({})'.format(missing_args[arg].get('help'), '/'.join(missing_args[arg].get('options')))
            value, cmd_value = interactive_input(arg, hint, cmd)
            cmd_arg_values[arg] = value

            # show applied params
            option = missing_args[arg].get('options')[0]
            param_str += '{} {} '.format(option, cmd_value)
    if param_str:
        logger.warning('Apply interactive input arguments: %s', param_str.strip())

    return cmd_arg_values


def validate_source_resource_id(cmd, namespace):
    '''Validate resource id of a source resource
    '''
    if getattr(namespace, 'source_id', None):
        if not is_valid_resource_id(namespace.source_id):
            e = InvalidArgumentValueError('Resource id is invalid: {}'.format(namespace.source_id))
            telemetry.set_exception(e, 'source-id-invalid')
            raise e

        source = get_source_resource_name(cmd)

        # For Web App, match slot pattern first:
        if source == RESOURCE.WebApp or source == RESOURCE.FunctionApp:
            slotPattern = WEB_APP_SLOT_RESOURCE
            matched = re.match(get_resource_regex(slotPattern), namespace.source_id, re.IGNORECASE)
            if matched:
                namespace.source_id = matched.group()
                return True
        if source == RESOURCE.SpringCloud:
            deploymentPattern = SPRING_APP_DEPLOYMENT_RESOURCE
            matched = re.match(get_resource_regex(deploymentPattern), namespace.source_id, re.IGNORECASE)
            if matched:
                namespace.source_id = matched.group()
                return True

        # For other source and Web App which cannot match slot pattern
        pattern = SOURCE_RESOURCES.get(source)
        matched = re.match(get_resource_regex(pattern),
                           namespace.source_id, re.IGNORECASE)
        if matched:
            namespace.source_id = matched.group()
            return True
        e = InvalidArgumentValueError(
            'Unsupported source resource id: {}. '
            'Source id pattern should be: {}'.format(namespace.source_id, pattern))
        telemetry.set_exception(e, 'source-id-unsupported')
        raise e
    return False


def validate_connection_id(namespace):
    '''Validate resource id of a connection
    '''
    if getattr(namespace, 'indentifier', None):
        matched = False
        for resource in list(SOURCE_RESOURCES.values()) + [WEB_APP_SLOT_RESOURCE, SPRING_APP_DEPLOYMENT_RESOURCE]:
            regex = '({})/providers/Microsoft.ServiceLinker/linkers/([^/]*)'.format(get_resource_regex(resource))
            matched = re.match(regex, namespace.indentifier, re.IGNORECASE)
            if matched:
                namespace.source_id = matched.group(1)
                namespace.connection_name = matched.group(2)
                return True
        if not matched:
            e = InvalidArgumentValueError('Connection id is invalid: {}'.format(namespace.indentifier))
            telemetry.set_exception(e, 'connection-id-invalid')
            raise e

    return False


def validate_target_resource_id(cmd, namespace):
    '''Validate resource id of a target resource
    '''
    if getattr(namespace, 'target_id', None):
        target = get_target_resource_name(cmd)
        if not (target == RESOURCE.FabricSql) and not is_valid_resource_id(namespace.target_id):
            e = InvalidArgumentValueError('Resource id is invalid: {}'.format(namespace.target_id))
            telemetry.set_exception(e, 'target-id-invalid')
            raise e
        pattern = TARGET_RESOURCES.get(target)
        matched = re.match(get_resource_regex(pattern), namespace.target_id, re.IGNORECASE)
        if matched:
            namespace.target_id = matched.group()
            return True
        e = InvalidArgumentValueError('Target resource id is invalid: {}. '
                                      'Target id pattern should be: {}'.format(namespace.target_id, pattern))
        telemetry.set_exception(e, 'target-id-unsupported')
        raise e

    return False


def validate_opt_out_auth_and_config(namespace):
    '''Validate if config and auth are both opted out
    '''
    opt_out_list = getattr(namespace, 'opt_out_list', None)
    if opt_out_list is not None and \
            OPT_OUT_OPTION.AUTHENTICATION.value in opt_out_list and \
            OPT_OUT_OPTION.CONFIGURATION_INFO.value in opt_out_list:
        return True
    return False


def get_missing_source_args(cmd, namespace):
    '''Get source resource related args
    '''
    source = get_source_resource_name(cmd)
    missing_args = {}

    for arg, content in SOURCE_RESOURCES_PARAMS.get(source, {}).items():
        missing_args[arg] = content

    # For WebApp, slot may needed
    args = SOURCE_RESOURCES_OPTIONAL_PARAMS.get(source)
    if args:
        for arg, content in args.items():
            if getattr(namespace, arg, None):
                missing_args[arg] = content
    return missing_args


def get_missing_source_create_args(cmd, namespace):
    '''Get source resource related args in create
    '''
    source = get_source_resource_name(cmd)
    missing_args = {}

    args = SOURCE_RESOURCES_CREATE_PARAMS.get(source)
    if args:
        for arg, content in args.items():
            if not getattr(namespace, arg, None):
                missing_args[arg] = content

    return missing_args


def get_missing_target_args(cmd):
    '''Get target resource related args
    '''
    target = get_target_resource_name(cmd)
    missing_args = {}

    if target in TARGET_RESOURCES_PARAMS:
        for arg, content in TARGET_RESOURCES_PARAMS.get(target).items():
            missing_args[arg] = content

    return missing_args


def get_missing_auth_args(cmd, namespace):
    '''Get auth info related args user didn't provide in command line
    '''
    source = get_source_resource_name(cmd)
    target = get_target_resource_name(cmd)
    missing_args = {}

    # check if there are auth_info related params
    auth_param_exist = False
    for _, params in AUTH_TYPE_PARAMS.items():
        for arg in params:
            if getattr(namespace, arg, None):
                auth_param_exist = True
                break

    if target == RESOURCE.ConfluentKafka:
        return missing_args
    # when keyvault csi is enabled, auth_type is userIdentity without subs_id and client_id
    if source == RESOURCE.KubernetesCluster and target == RESOURCE.KeyVault:
        if getattr(namespace, 'enable_csi', None):
            if auth_param_exist:
                logger.warning('When CSI driver is enabled (--enable-csi), \
                               Service Connector uses the user assigned managed identity generated by AKS \
                               azure-keyvault-secrets-provider add-on to authenticate. \
                               Additional auth info is ignored.')
            setattr(namespace, 'user_identity_auth_info', {
                'auth_type': 'userAssignedIdentity'
            })
            return missing_args
        if not auth_param_exist:
            setattr(namespace, 'enable_csi', True)
            setattr(namespace, 'user_identity_auth_info', {
                'auth_type': 'userAssignedIdentity'
            })
            logger.warning('Auth info is not specified, use secrets store csi driver as default: --enable-csi')
            return missing_args

    # ACA as target use null auth
    if target == RESOURCE.ContainerApp:
        return missing_args

    if source and target and not auth_param_exist:
        default_auth_type = SUPPORTED_AUTH_TYPE.get(source, {}).get(target, {})[0]

        for arg, content in AUTH_TYPE_PARAMS.get(default_auth_type).items():
            if getattr(namespace, arg, None) is None:
                missing_args[arg] = content

    return missing_args


def get_missing_connection_name(namespace):
    '''Get connection_name arg if user didn't provide it in command line
    '''
    missing_args = {}
    if getattr(namespace, 'connection_name', None) is None:
        missing_args['connection_name'] = {
            'help': 'The connection name',
            'options': ['--connection']
        }

    return missing_args


def get_missing_client_type(namespace):
    '''Get client_type arg if user didn't provide it in command line
    '''
    missing_args = {}
    if getattr(namespace, 'client_type', None) is None:
        missing_args['client_type'] = {
            'help': 'Client type of the connection',
            'options': ['--client-type']
        }

    return missing_args


def validate_local_default_params(cmd, namespace):  # pylint: disable=unused-argument
    '''Get missing args of local connection command
    '''
    missing_args = {}

    if getattr(namespace, 'id', None):
        namespace.id = namespace.id.lower()
        if not is_valid_resource_id(namespace.id):
            raise InvalidArgumentValueError(
                'Resource id is invalid: {}'.format(namespace.id))
        resource = LOCAL_CONNECTION_RESOURCE.lower()
        for item in re.findall(r'(\{[^\{\}]*\})', resource):
            resource = resource.replace(item, '([^/]*)')

        matched = re.match(resource, namespace.id)
        if matched:
            namespace.resource_group_name = matched.group(2)
            namespace.location = matched.group(3)
            namespace.connection_name = matched.group(4)
        else:
            raise InvalidArgumentValueError(
                'Unsupported resource id: {}'.format(namespace.id))
    else:
        if not getattr(namespace, 'resource_group_name', None):
            missing_args.update(
                {'resource_group_name': LOCAL_CONNECTION_PARAMS.get("resource_group_name")})
    return missing_args


def apply_local_default_params(cmd, namespace, arg_values):  # pylint: disable=unused-argument
    for arg in LOCAL_CONNECTION_PARAMS:
        if arg in arg_values:
            setattr(namespace, arg, arg_values.get(arg, None))


def validate_local_list_params(cmd, namespace):  # pylint: disable=unused-argument
    missing_args = {}
    if getattr(namespace, 'resource_group', None) is None:
        missing_args.update(LOCAL_CONNECTION_PARAMS.get("resource_group"))
    return missing_args


def validate_list_params(cmd, namespace):
    '''Get missing args of list command
    '''
    missing_args = {}
    if not validate_source_resource_id(cmd, namespace):
        missing_args.update(get_missing_source_args(cmd, namespace))
    return missing_args


def validate_create_params(cmd, namespace):
    '''Get missing args of create command
    '''
    missing_args = {}
    if not validate_source_resource_id(cmd, namespace):
        missing_args.update(get_missing_source_args(cmd, namespace))
    missing_args.update(get_missing_source_create_args(cmd, namespace))
    if not validate_target_resource_id(cmd, namespace):
        missing_args.update(get_missing_target_args(cmd))
    if not validate_opt_out_auth_and_config(namespace):
        missing_args.update(get_missing_auth_args(cmd, namespace))
    return missing_args


def validate_local_create_params(cmd, namespace):
    '''Get missing args of create command
    '''
    missing_args = {}

    if not validate_target_resource_id(cmd, namespace):
        missing_args.update(get_missing_target_args(cmd))
    missing_args.update(get_missing_auth_args(cmd, namespace))
    return missing_args


def validate_addon_params(cmd, namespace):
    '''Get missing args of add command with '--new'
    '''
    missing_args = {}
    if not validate_source_resource_id(cmd, namespace):
        missing_args.update(get_missing_source_args(cmd, namespace))
    missing_args.update(get_missing_auth_args(cmd, namespace))
    return missing_args


def validate_update_params(cmd, namespace):
    '''Get missing args of update command
    '''
    missing_args = {}
    if not validate_connection_id(namespace) and not validate_source_resource_id(cmd, namespace):
        missing_args.update(get_missing_source_args(cmd, namespace))
    # missing_args.update(get_missing_auth_args(cmd, namespace))
    missing_args.update(get_missing_connection_name(namespace))
    return missing_args


def validate_local_update_params(cmd, namespace):  # pylint: disable=unused-argument
    '''Get missing args of update command
    '''
    missing_args = {}
    # missing_args.update(get_missing_auth_args(cmd, namespace))
    return missing_args


def validate_default_params(cmd, namespace):
    '''Get missing args of commands except for list, create
    '''
    missing_args = {}
    if not validate_connection_id(namespace):
        missing_args.update(get_missing_source_args(cmd, namespace))
    missing_args.update(get_missing_connection_name(namespace))
    return missing_args


def validate_connection_name(name):
    if not re.match(r'^[A-Za-z0-9\._]+$', name):
        e = InvalidArgumentValueError("Resource name can only contain letters (A-Z, a-z), "
                                      "numbers (0-9), periods ('.'), and underscores ('_')")
        telemetry.set_exception('connection-name-invalid')
        raise e
    return True


def apply_source_args(cmd, namespace, arg_values):
    '''Set source resource id by arg_values
    '''
    source = get_source_resource_name(cmd)
    resource = SOURCE_RESOURCES.get(source)
    if check_required_args(resource, arg_values):
        namespace.source_id = resource.format(
            subscription=get_subscription_id(cmd.cli_ctx),
            **arg_values
        )
    apply_source_optional_args(cmd, namespace, arg_values)


def apply_source_optional_args(cmd, namespace, arg_values):
    '''Set source resource id by optional arg_values
    '''
    source = get_source_resource_name(cmd)
    if source == RESOURCE.WebApp or source == RESOURCE.FunctionApp:
        if arg_values.get('slot', None):
            resource = WEB_APP_SLOT_RESOURCE
            if check_required_args(resource, arg_values):
                namespace.source_id = resource.format(
                    subscription=get_subscription_id(cmd.cli_ctx),
                    **arg_values
                )
    if source == RESOURCE.SpringCloud:
        if arg_values.get('deployment', None):
            resource = SPRING_APP_DEPLOYMENT_RESOURCE
            if check_required_args(resource, arg_values):
                namespace.source_id = resource.format(
                    subscription=get_subscription_id(cmd.cli_ctx),
                    **arg_values
                )


def apply_source_create_args(cmd, namespace, arg_values):
    '''Set source resource related args in create by arg_values
    '''
    source = get_source_resource_name(cmd)
    for arg in SOURCE_RESOURCES_CREATE_PARAMS.get(source, {}):
        if arg in arg_values:
            setattr(namespace, arg, arg_values.get(arg, None))


def apply_target_args(cmd, namespace, arg_values):
    '''Set target resource id by arg_values
    '''
    target = get_target_resource_name(cmd)
    resource = TARGET_RESOURCES.get(target)
    if check_required_args(resource, arg_values):
        namespace.target_id = resource.format(
            subscription=get_subscription_id(cmd.cli_ctx),
            **arg_values
        )


def apply_auth_args(cmd, namespace, arg_values):
    '''Set auth info by arg_values
    '''
    source = get_source_resource_name(cmd)
    target = get_target_resource_name(cmd)
    if source and target:
        auth_types = SUPPORTED_AUTH_TYPE.get(source, {}).get(target, {})
        for auth_type in auth_types:
            if auth_type == AUTH_TYPE.Null:
                continue
            for arg in AUTH_TYPE_PARAMS.get(auth_type):
                if arg in arg_values:
                    setattr(namespace, arg, arg_values.get(arg, None))
                    if arg == 'workload_identity_auth_info':
                        apply_workload_identity(namespace, arg_values)


def apply_workload_identity(namespace, arg_values):
    output = run_cli_cmd('az identity show --ids "{}"'.format(
        arg_values.get('workload_identity_auth_info')
    ))
    if output:
        client_id = output.get('clientId')
        subs_id = arg_values.get('workload_identity_auth_info').split('/')[2]
    else:
        raise ValidationError('Invalid user identity resource ID for workload identity.')
    setattr(namespace, 'user_identity_auth_info',
            {
                'client_id': client_id,
                'subscription_id': subs_id,
                'auth_type': 'userAssignedIdentity'
            })


def apply_connection_name(namespace, arg_values):
    '''Set connection_name by arg_values
    '''
    if getattr(namespace, 'connection_name', None) is None:
        namespace.connection_name = arg_values.get('connection_name', None)


def apply_client_type(namespace, arg_values):
    '''Set client_type by arg_values
    '''
    if getattr(namespace, 'client_type', None) is None:
        namespace.client_type = arg_values.get('client_type', None)


def apply_list_params(cmd, namespace, arg_values):
    '''Set list command missing args
    '''
    apply_source_args(cmd, namespace, arg_values)


def apply_create_params(cmd, namespace, arg_values):
    '''Set create command missing args
    '''
    apply_source_args(cmd, namespace, arg_values)
    apply_source_create_args(cmd, namespace, arg_values)
    apply_target_args(cmd, namespace, arg_values)
    apply_auth_args(cmd, namespace, arg_values)


def apply_local_create_params(cmd, namespace, arg_values):
    '''Set create command missing args
    '''
    apply_target_args(cmd, namespace, arg_values)
    apply_auth_args(cmd, namespace, arg_values)


def apply_addon_params(cmd, namespace, arg_values):
    '''Set addon command missing args
    '''
    apply_source_args(cmd, namespace, arg_values)
    apply_auth_args(cmd, namespace, arg_values)


def apply_update_params(cmd, namespace, arg_values):
    '''Set update command missing args
    '''
    apply_source_args(cmd, namespace, arg_values)
    apply_connection_name(namespace, arg_values)
    apply_auth_args(cmd, namespace, arg_values)


def apply_local_update_params(cmd, namespace, arg_values):
    '''Set update command missing args
    '''
    apply_auth_args(cmd, namespace, arg_values)


def apply_default_params(cmd, namespace, arg_values):
    '''Set missing args of commands except for list, create
    '''
    apply_source_args(cmd, namespace, arg_values)
    apply_connection_name(namespace, arg_values)


def validate_local_params(cmd, namespace):
    '''Validate command parameters
    '''
    def _validate_and_apply(validate, apply):
        missing_args = validate(cmd, namespace)
        if missing_args:
            arg_values = intelligent_experience(cmd, namespace, missing_args)
            apply(cmd, namespace, arg_values)
    # for command: 'list'
    if cmd.name.endswith(' list'):
        _validate_and_apply(validate_local_list_params,
                            apply_local_default_params)
    else:
        _validate_and_apply(validate_local_default_params,
                            apply_local_default_params)

    # for command: 'create'
    if 'create' in cmd.name:
        # if --new is specified
        if getattr(namespace, 'connection_name', None) is None:
            namespace.connection_name = generate_connection_name(cmd)
        else:
            validate_connection_name(namespace.connection_name)
        if getattr(namespace, 'new_addon', None):
            _validate_and_apply(validate_addon_params, apply_addon_params)
        else:
            _validate_and_apply(validate_local_create_params,
                                apply_local_create_params)
        if getattr(namespace, 'client_type', None) is None:
            namespace.client_type = get_client_type(cmd, namespace)
    # for command: update
    elif 'update' in cmd.name:
        _validate_and_apply(validate_local_update_params,
                            apply_local_update_params)


def validate_params(cmd, namespace):
    '''Validate command parameters
    '''
    def _validate_and_apply(validate, apply):
        missing_args = validate(cmd, namespace)
        if missing_args:
            arg_values = intelligent_experience(cmd, namespace, missing_args)
            apply(cmd, namespace, arg_values)

    # for command: 'list'
    if cmd.name.endswith(' list'):
        _validate_and_apply(validate_list_params, apply_list_params)
    # for command: 'create'
    elif 'create' in cmd.name:
        # if --new is specified
        if getattr(namespace, 'connection_name', None) is None:
            namespace.connection_name = generate_connection_name(cmd)
        else:
            validate_connection_name(namespace.connection_name)
        if getattr(namespace, 'new_addon', None):
            _validate_and_apply(validate_addon_params, apply_addon_params)
        else:
            _validate_and_apply(validate_create_params, apply_create_params)
        if getattr(namespace, 'client_type', None) is None:
            namespace.client_type = get_client_type(cmd, namespace)
    # for command: update
    elif 'update' in cmd.name:
        _validate_and_apply(validate_update_params, apply_update_params)
    # for command: all others
    else:
        _validate_and_apply(validate_default_params, apply_default_params)


def validate_kafka_params(cmd, namespace):
    def _validate_and_apply(validate, apply):
        missing_args = validate(cmd, namespace)
        if missing_args:
            arg_values = intelligent_experience(cmd, namespace, missing_args)
            apply(cmd, namespace, arg_values)

    _validate_and_apply(validate_list_params, apply_list_params)
    if 'create {}'.format(RESOURCE.ConfluentKafka.value) in cmd.name:
        if getattr(namespace, 'connection_name', None) is None:
            namespace.connection_name = generate_connection_name(cmd)
        elif namespace.connection_name.endswith('_schema'):
            raise InvalidArgumentValueError("Connection name of {} can not end with "
                                            "'_schema'".format(RESOURCE.ConfluentKafka.value))
        else:
            validate_connection_name(namespace.connection_name)

        if getattr(namespace, 'client_type', None) is None:
            namespace.client_type = get_client_type(cmd, namespace)


def validate_connstr_props(cmd, namespace):
    if 'create {}'.format(RESOURCE.FabricSql.value) in cmd.name:
        if getattr(namespace, 'connstr_props', None) is None:
            namespace.connstr_props = generate_fabric_connstr_props(namespace.target_id)

            if namespace.connstr_props is None:
                e = InvalidArgumentValueError("Fabric Connection String Properties must be provided")
                telemetry.set_exception('fabric-connstr-props-unavailable')
                raise e
        else:
            fabric_server = namespace.connstr_props.get('Server') or namespace.connstr_props.get('Data Source')
            fabric_database = namespace.connstr_props.get('Database') or namespace.connstr_props.get('Initial Catalog')

            if not fabric_server or not fabric_database:
                e = InvalidArgumentValueError("Fabric Connection String Properties must contain Server and Database")
                telemetry.set_exception('fabric-connstr-props-invalid')
                raise e


def validate_service_state(linker_parameters):
    '''Validate whether user provided params are applicable to service state
    '''
    target_type = linker_parameters.get('target_service', {}).get('type')

    # AzureResource and other types (e.g., FabricResource, SelfHostedResource)
    if target_type == "AzureResource":
        target_id = linker_parameters.get('target_service', {}).get('id')
    else:
        target_id = linker_parameters.get('target_service', {}).get('endpoint')

    for target, resource_id in TARGET_RESOURCES.items():
        matched = re.match(get_resource_regex(resource_id), target_id, re.IGNORECASE)
        if matched:
            target_type = target

    if target_type == RESOURCE.AppConfig and linker_parameters.get('auth_info', {}).get('auth_type') == 'secret':
        segments = parse_resource_id(target_id)
        rg = segments.get('resource_group')
        name = segments.get('name')
        sub = segments.get('subscription')
        if not rg or not name:
            return

        output = run_cli_cmd('az appconfig show -g "{}" -n "{}" --subscription "{}"'.format(rg, name, sub))
        if output and output.get('disableLocalAuth') is True:
            raise ValidationError('Secret as auth type is not allowed when local auth is disabled for the '
                                  'specified appconfig, you may use service principal or managed identity.')

    if target_type == RESOURCE.Redis:
        auth_type = linker_parameters.get('auth_info', {}).get('auth_type')
        if auth_type == AUTH_TYPE.Secret.value or auth_type == AUTH_TYPE.SecretAuto.value:
            return
        redis = run_cli_cmd('az redis show --ids "{}"'.format(target_id))
        if redis.get('redisConfiguration', {}).get('aadEnabled', 'False') != "True":
            raise ValidationError('Please enable Microsoft Entra Authentication on your Redis first. '
                                  'Note that it will cause your cache instances to reboot to load new '
                                  'configuration and result in a failover. Consider performing the '
                                  'operation during low traffic or outside of business hours.')


def get_default_object_id_of_current_user(cmd, namespace):  # pylint: disable=unused-argument
    user_account_auth_info = getattr(namespace, 'user_account_auth_info', None)
    if user_account_auth_info and not user_account_auth_info.get('principal_id', None):
        user_object_id = get_object_id_of_current_user()
        user_account_auth_info['principal_id'] = user_object_id
        setattr(namespace, 'user_account_auth_info', user_account_auth_info)



# coding=utf-8
# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.help_files import helps  # pylint: disable=unused-import

# pylint: disable=line-too-long, too-many-lines


helps['postgres flexible-server'] = """
type: group
short-summary: Manage Azure Database for PostgreSQL Flexible Servers.
"""

helps['postgres flexible-server create'] = """
type: command
short-summary: Create a PostgreSQL flexible server.
long-summary: >
    Create a PostgreSQL flexible server with custom or default configuration. For more information for network configuration, see

    - Configure public access

    https://learn.microsoft.com/en-us/azure/postgresql/flexible-server/how-to-manage-firewall-cli

    - Configure private access

    https://learn.microsoft.com/en-us/azure/postgresql/flexible-server/how-to-manage-virtual-network-cli

examples:
  - name: >
      Create a PostgreSQL flexible server with custom parameters
    text: >
        az postgres flexible-server create --location northeurope --resource-group testGroup \\
          --name testserver --admin-user username --admin-password password \\
          --sku-name Standard_D2s_v3 --tier GeneralPurpose --public-access 153.24.26.117 --storage-size 128 \\
          --tags "key=value" --version 17 --high-availability ZoneRedundant --zone 1 \\
          --standby-zone 3
  - name: >
      Create a PostgreSQL flexible server using Premium SSD v2 Disks.
    text: >
      # set storage type to "PremiumV2_LRS" and provide values for Storage size (in GiB), IOPS (operations/sec), and Throughput (MB/sec).

      az postgres flexible-server create --location northeurope --resource-group testGroup \\
          --name testserver --admin-user username --admin-password password \\
          --sku-name Standard_B1ms --tier Burstable --storage-type PremiumV2_LRS --storage-size 128 --iops 3000 --throughput 125
  - name: >
      Create a PostgreSQL flexible server with default parameters and public access enabled by default. \
      Resource group, server name, username, password, and default database will be created by CLI
    text: >
        az postgres flexible-server create
  - name: >
      Create a PostgreSQL flexible server with public access and add the range of IP address to have access to this server.
      The --public-access parameter can be 'Disabled', 'Enabled', 'All', 'None', <startIpAddress>, or <startIpAddress>-<endIpAddress>
    text: >
      az postgres flexible-server create --resource-group testGroup --name testserver --public-access 125.23.54.31-125.23.54.35
  - name: >
      Create a PostgreSQL flexible server with private access. If provided virtual network and subnet do not exists, virtual network and subnet with the specified address prefixes will be created.
    text: >
      az postgres flexible-server create --resource-group testGroup --name testserver --vnet myVnet --subnet mySubnet --address-prefixes 10.0.0.0/16 --subnet-prefixes 10.0.0.0/24
  - name: >
      Create a PostgreSQL flexible server using a new subnet resource ID and new private DNS zone resource ID. The subnet and DNS zone can be created in different subscription or resource group.
    text: |
      az postgres flexible-server create \\
        --resource-group testGroup --name testserver \\
        --subnet /subscriptions/{SubID}/resourceGroups/{ResourceGroup}/providers/Microsoft.Network/virtualNetworks/{VNetName}/subnets/{SubnetName} \\
        --private-dns-zone /subscriptions/{SubID}/resourceGroups/{resourceGroup}/providers/Microsoft.Network/privateDnsZones/testPostgreSQLFlexibleDnsZone.private.postgres.database.azure.com \\
        --address-prefixes 172.0.0.0/16 --subnet-prefixes 172.0.0.0/24
  - name: >
      Create a PostgreSQL flexible server using existing network resources in the same resource group.
      The provided subnet should not have any other resource deployed in it and this subnet will be delegated to Microsoft.DBforPostgreSQL/flexibleServers, if not already delegated.
      The private DNS zone will be linked to the virtual network if not already linked.
    text: >
      # create vnet

      az network vnet create --resource-group testGroup --name testVnet --location testLocation --address-prefixes 172.0.0.0/16


      # create subnet

      az network vnet subnet create --resource-group testGroup --vnet-name testVnet --address-prefixes 172.0.0.0/24 --name testSubnet


      # create private dns zone

      az network private-dns zone create -g testGroup -n testDNS.private.postgres.database.azure.com


      az postgres flexible-server create --resource-group testGroup \\
        --name testserver --location testLocation \\
        --subnet /subscriptions/{SubId}/resourceGroups/{testGroup}/providers/Microsoft.Network/virtualNetworks/tesetVnet/subnets/testSubnet \\
        --private-dns-zone /subscriptions/{SubId}/resourceGroups/{testGroup}/providers/Microsoft.Network/privateDnsZones/testDNS.postgres.database.azure.com\\


      az postgres flexible-server create --resource-group testGroup --name testserver \\
        --vnet testVnet --subnet testSubnet --location testLocation \\
        --private-dns-zone /subscriptions/{SubId}/resourceGroups/{testGroup}/providers/Microsoft.Network/privateDnsZones/testDNS.postgres.database.azure.com
  - name: >
      Create a PostgreSQL flexible server using existing network resources in the different resource group / subscription.
    text: >
      az postgres flexible-server create --resource-group testGroup \\
         --name testserver --location testLocation \\
        --subnet /subscriptions/{SubId2}/resourceGroups/{testGroup2}/providers/Microsoft.Network/virtualNetworks/tesetVnet/subnets/testSubnet \\
        --private-dns-zone /subscriptions/{SubId2}/resourceGroups/{testGroup2}/providers/Microsoft.Network/privateDnsZones/testDNS.postgres.database.azure.com
  - name: >
      Create a PostgreSQL flexible server with data encryption.
    text: >
      # create keyvault

      az keyvault create -g testGroup -n testVault --location testLocation \\
        --enable-purge-protection true


      # create key in keyvault and save its key identifier

      keyIdentifier=$(az keyvault key create --name testKey -p software \\
        --vault-name testVault --query key.kid -o tsv)


      # create identity and save its principalId

      identityPrincipalId=$(az identity create -g testGroup --name testIdentity \\
        --location testLocation --query principalId -o tsv)


      # add testIdentity as an access policy with key permissions 'Wrap Key', 'Unwrap Key', 'Get' and 'List' inside testVault

      az keyvault set-policy -g testGroup -n testVault --object-id $identityPrincipalId \\
        --key-permissions wrapKey unwrapKey get list


      # create flexible server with data encryption enabled

      az postgres flexible-server create -g testGroup -n testServer --location testLocation \\
        --key $keyIdentifier --identity testIdentity
  - name: >
      Create a PostgreSQL flexible server with Microsoft Entra auth as well as password auth.
    text: >
      # create flexible server with aad auth and password auth enabled

      az postgres flexible-server create -g testGroup -n testServer --location testLocation \\
        --microsoft-entra-auth Enabled
  - name: >
      Create a PostgreSQL flexible server with Microsoft Entra auth only and primary administrator specified.
    text: >
      # create flexible server with aad only auth and password auth disabled with primary administrator specified

      az postgres flexible-server create -g testGroup -n testServer --location testLocation \\
        --microsoft-entra-auth Enabled --password-auth Disabled \\
        --admin-object-id 00000000-0000-0000-0000-000000000000 --admin-display-name john@contoso.com --admin-type User
  - name: >
      Create a PostgreSQL flexible server with public access, geo-redundant backup enabled and add the range of IP address to have access to this server.
      The --public-access parameter can be 'All', 'None', <startIpAddress>, or <startIpAddress>-<endIpAddress>
    text: >
      az postgres flexible-server create --resource-group testGroup --name testserver --geo-redundant-backup Enabled --public-access 125.23.54.31-125.23.54.35
  - name: >
      Create a PostgreSQL flexible server with data encryption for geo-rundundant backup enabled server.
    text: >
      # create keyvault

      az keyvault create -g testGroup -n testVault --location testLocation \\
        --enable-purge-protection true


      # create key in keyvault and save its key identifier

      keyIdentifier=$(az keyvault key create --name testKey -p software \\
        --vault-name testVault --query key.kid -o tsv)


      # create identity and save its principalId

      identityPrincipalId=$(az identity create -g testGroup --name testIdentity \\
        --location testLocation --query principalId -o tsv)


      # add testIdentity as an access policy with key permissions 'Wrap Key', 'Unwrap Key', 'Get' and 'List' inside testVault

      az keyvault set-policy -g testGroup -n testVault --object-id $identityPrincipalId \\
        --key-permissions wrapKey unwrapKey get list

      # create keyvault in geo-paired region

      az keyvault create -g testGroup -n geoVault --location geoPairedLocation \\
        --enable-purge-protection true


      # create key in keyvault and save its key identifier

      geoKeyIdentifier=$(az keyvault key create --name geoKey -p software \\
        --vault-name geoVault --query key.kid -o tsv)


      # create identity in geo-raired location and save its principalId

      geoIdentityPrincipalId=$(az identity create -g testGroup --name geoIdentity \\
        --location geoPairedLocation --query principalId -o tsv)


      # add testIdentity as an access policy with key permissions 'Wrap Key', 'Unwrap Key', 'Get' and 'List' inside testVault

      az keyvault set-policy -g testGroup -n geoVault --object-id $geoIdentityPrincipalId \\
        --key-permissions wrapKey unwrapKey get list


      # create flexible server with data encryption enabled for geo-backup Enabled server

      az postgres flexible-server create -g testGroup -n testServer --location testLocation --geo-redundant-backup Enabled \\
        --key $keyIdentifier --identity testIdentity --backup-key $geoKeyIdentifier --backup-identity geoIdentity

  - name: >
      Create flexible server with custom storage performance tier. Accepted values "P4", "P6", "P10", "P15", "P20", "P30", \\
      "P40", "P50", "P60", "P70", "P80". Actual allowed values depend on the --storage-size selection for flexible server creation. \\
      Default value for storage performance tier depends on the --storage-size selected for flexible server creation.
    text: >
      az postgres flexible-server create -g testGroup -n testServer --location testLocation --performance-tier P15

  - name: >
      Create flexible server with storage auto-grow as Enabled. Accepted values Enabled / Disabled. Default value for storage auto-grow is "Disabled".
    text: >
      az postgres flexible-server create -g testGroup -n testServer --location testLocation --storage-auto-grow Enabled

  - name: >
      Create elastic cluster with node count of 5. Default node count is 2 when --cluster-option is "ElasticCluster".
    text: >
      az postgres flexible-server create -g testGroup -n testCluster --location testLocation --cluster-option ElasticCluster --node-count 5
"""

helps['postgres flexible-server show'] = """
type: command
short-summary: Get the details of a flexible server.
examples:
  - name: Get the details of a flexible server
    text: az postgres flexible-server show --resource-group testGroup --name testserver
"""

helps['postgres flexible-server list'] = """
type: command
short-summary: List available flexible servers.
examples:
  - name: List all PostgreSQL flexible servers in a subscription.
    text: az postgres flexible-server list
  - name: List all PostgreSQL flexible servers in a resource group.
    text: az postgres flexible-server list --resource-group testGroup
  - name: List all PostgreSQL flexible servers in a resource group in table format.
    text: az postgres flexible-server list --resource-group testGroup --output table
"""

helps['postgres flexible-server update'] = """
type: command
short-summary: Update a flexible server.
examples:
  - name: Update a flexible server's sku, using local context for server and resource group.
    text: az postgres flexible-server update --sku-name Standard_D4s_v3
  - name: Update a server's tags.
    text: az postgres flexible-server update --resource-group testGroup --name testserver --tags "k1=v1" "k2=v2"
  - name: Reset password
    text: az postgres flexible-server update --resource-group testGroup --name testserver -p password123
  - name: Update a flexible server to enable Microsoft Entra auth for password auth enabled server
    text: az postgres flexible-server update --resource-group testGroup --name testserver --microsoft-entra-auth Enabled
  - name: Change key/identity for data encryption. Data encryption cannot be enabled post server creation, this will only update the key/identity.
    text: >
      # get key identifier of the existing key

      newKeyIdentifier=$(az keyvault key show --vault-name testVault --name testKey \\
        --query key.kid -o tsv)


      # update server with new key/identity

      az postgres flexible-server update --resource-group testGroup --name testserver \\
        --key $newKeyIdentifier --identity newIdentity
  - name: Update a flexible server to update private DNS zone for a VNET enabled server, using private DNS zone in the same resource group and subscription. Private DNS zone will be created Private DNS zone will be linked to the VNET if not already linked.
    text: az postgres flexible-server update --resource-group testGroup --name testserver --private-dns-zone testDNS2.postgres.database.azure.com
  - name: Update a flexible server to update private DNS zone for a VNET enabled server, using private DNS zone in the different resource group and subscription. Private DNS zone will be linked to the VNET if not already linked.
    text: az postgres flexible-server update --resource-group testGroup --name testserver --private-dns-zone /subscriptions/{SubId2}/resourceGroups/{testGroup2}/providers/Microsoft.Network/privateDnsZones/testDNS.postgres.database.azure.com
  - name: Update a flexible server's storage to enable / disable storage auto-grow.
    text: az postgres flexible-server update --resource-group testGroup --name testserver --storage-auto-grow Enabled
  - name: Update a flexible server's storage to set custom storage performance tier.
    text: az postgres flexible-server update --resource-group testGroup --name testserver --performance-tier P15
  - name: Update a flexible server's storage to set IOPS (operations/sec). Server must be using Premium SSD v2 Disks.
    text: az postgres flexible-server update --resource-group testGroup --name testserver --iops 3000
  - name: Update a flexible server's storage to set Throughput (MB/sec). Server must be using Premium SSD v2 Disks.
    text: az postgres flexible-server update --resource-group testGroup --name testserver --throughput 125
  - name: Update a flexible server's cluster size by scaling up node count. Must be an Elastic Cluster.
    text: az postgres flexible-server update --resource-group testGroup --name testcluster --node-count 6
"""

helps['postgres flexible-server restore'] = """
type: command
short-summary: Restore a flexible server from backup.
examples:
  - name: Restore 'testserver' to a specific point-in-time as a new server 'testserverNew'.
    text: az postgres flexible-server restore --resource-group testGroup --name testserverNew --source-server testserver --restore-time "2017-06-15T13:10:00Z"
  - name: Restore 'testserver' to current point-in-time as a new server 'testserverNew'.
    text: az postgres flexible-server restore --resource-group testGroup --name testserverNew --source-server testserver
  - name: >
      Restore 'testserver' to current point-in-time as a new server 'testserverNew' in a different resource group. \\
      Here --resource-group is for the target server's resource group, and --source-server must be passed as resource ID.
    text: >
      az postgres flexible-server restore --resource-group testGroup --name testserverNew \\
        --source-server /subscriptions/{testSubscription}/resourceGroups/{sourceResourceGroup}/providers/Microsoft.DBforPostgreSQL/flexibleServers/{sourceServerName}
  - name: >
      Restore 'testserver' to current point-in-time as a new server 'testserverNew' in a different subscription. \\
      Here --resource-group is for the target server's resource group, and --source-server must be passed as resource ID. \\
      This resource ID can be in a subscription different than the subscription used for az account set.
    text: >
      az postgres flexible-server restore --resource-group testGroup --name testserverNew \\
        --source-server /subscriptions/{sourceSubscriptionId}/resourceGroups/{sourceResourceGroup}/providers/Microsoft.DBforPostgreSQL/flexibleServers/{sourceServerName}
  - name: Restore 'testserver' to current point-in-time as a new server 'testserverNew' using Premium SSD v2 Disks by setting storage type to "PremiumV2_LRS"
    text: az postgres flexible-server restore --resource-group testGroup --name testserverNew --source-server testserver --storage-type PremiumV2_LRS
"""

helps['postgres flexible-server restart'] = """
type: command
short-summary: Restart a flexible server.
examples:
  - name: Restart a flexible server.
    text: az postgres flexible-server restart --resource-group testGroup --name testserver
  - name: Restart a server with planned failover
    text: az postgres flexible-server restart --resource-group testGroup --name testserver --failover Planned
  - name: Restart a server with forced failover
    text: az postgres flexible-server restart --resource-group testGroup --name testserver --failover Forced
"""

helps['postgres flexible-server start'] = """
type: command
short-summary: Start a flexible server.
examples:
  - name: Start a flexible server.
    text: az postgres flexible-server start --resource-group testGroup --name testserver
"""

helps['postgres flexible-server stop'] = """
type: command
short-summary: Stop a flexible server.
examples:
  - name: Stop a flexible server.
    text: az postgres flexible-server stop --resource-group testGroup --name testserver
"""

helps['postgres flexible-server wait'] = """
type: command
short-summary: Wait for the flexible server to satisfy certain conditions.
example:
  - name: Wait for the flexible server to satisfy certain conditions.
    text: az postgres flexible-server wait --exists --resource-group testGroup --name testserver
"""

helps['postgres flexible-server delete'] = """
type: command
short-summary: Delete a flexible server.
examples:
  - name: Delete a flexible server.
    text: az postgres flexible-server delete --resource-group testGroup --name testserver
  - name: Delete a flexible server without prompt or confirmation.
    text: az postgres flexible-server delete --resource-group testGroup --name testserver --yes
"""

helps['postgres flexible-server db'] = """
type: group
short-summary: Manage PostgreSQL databases on a flexible server.
"""

helps['postgres flexible-server db create'] = """
type: command
short-summary: Create a PostgreSQL database on a flexible server.
examples:
  - name: Create database 'testDatabase' in the flexible server 'testserver' with the default parameters.
    text: az postgres flexible-server db create --resource-group testGroup --server-name testserver --database-name testDatabase
  - name: Create database 'testDatabase' in the flexible server 'testserver' with a given character set and collation rules.
    text: az postgres flexible-server db create --resource-group testGroup --server-name testserver --database-name testDatabase \\
            --charset validCharset --collation validCollation
"""

helps['postgres flexible-server db delete'] = """
type: command
short-summary: Delete a database on a flexible server.
examples:
  - name: Delete database 'testDatabase' in the flexible server 'testserver'.
    text: az postgres flexible-server db delete --resource-group testGroup --server-name testserver --database-name testDatabase
"""

helps['postgres flexible-server db list'] = """
type: command
short-summary: List the databases for a flexible server.
examples:
  - name: List databases in the flexible server 'testserver'.
    text: az postgres flexible-server db list --resource-group testGroup --server-name testserver
  - name: List databases in the flexible server 'testserver' in table format.
    text: az postgres flexible-server db list --resource-group testGroup --server-name testserver --output table
"""

helps['postgres flexible-server db show'] = """
type: command
short-summary: Show the details of a database.
examples:
  - name: Show database 'testDatabase' in the server 'testserver'.
    text: az postgres flexible-server db show --resource-group testGroup --server-name testserver --database-name testDatabase
"""

helps['postgres flexible-server firewall-rule'] = """
type: group
short-summary: Manage firewall rules for a server.
"""

helps['postgres flexible-server firewall-rule create'] = """
type: command
short-summary: Create a new firewall rule for a flexible server.
examples:
  - name: >
      Create a firewall rule allowing connections from a specific IP address.
    text: >
      az postgres flexible-server firewall-rule create --resource-group testGroup --name testserver --rule-name allowip --start-ip-address 107.46.14.221
  - name: >
      Create a firewall rule allowing connections from an IP address range.
    text: >
        az postgres flexible-server firewall-rule create --resource-group testGroup --name testserver --rule-name allowiprange --start-ip-address 107.46.14.0 --end-ip-address 107.46.14.221
  - name: >
      Create a firewall rule allowing connections to all Azure services
    text: >
      az postgres flexible-server firewall-rule create --resource-group testGroup --name testserver --rule-name allowazureservices --start-ip-address 0.0.0.0
"""

helps['postgres flexible-server firewall-rule list'] = """
type: command
short-summary: List all firewall rules for a flexible server.
example:
  - name: List all firewall rules for a server.
    text: az postgres flexible-server firewall-rule list --resource-group testGroup --name testserver
  - name: List all firewall rules for a server in table format.
    text: az postgres flexible-server firewall-rule list --resource-group testGroup --name testserver --output table
"""

helps['postgres flexible-server firewall-rule show'] = """
type: command
short-summary: Get the details of a firewall rule.
examples:
  - name: Get the details of a firewall rule.
    text: az postgres flexible-server firewall-rule show --rule-name testRule --resource-group testGroup --name testserver
"""

helps['postgres flexible-server firewall-rule update'] = """
type: command
short-summary: Update a firewall rule.
examples:
  - name: Update a firewall rule's start IP address.
    text: az postgres flexible-server firewall-rule update --resource-group testGroup --name testserver --rule-name allowiprange --start-ip-address 107.46.14.1
  - name: Update a firewall rule's start and end IP address.
    text: az postgres flexible-server firewall-rule update --resource-group testGroup --name testserver --rule-name allowiprange --start-ip-address 107.46.14.2 --end-ip-address 107.46.14.218
"""

helps['postgres flexible-server firewall-rule delete'] = """
type: command
short-summary: Delete a firewall rule.
examples:
  - name: Delete a firewall rule.
    text: az postgres flexible-server firewall-rule delete --rule-name testRule --resource-group testGroup --name testserver
"""

helps['postgres flexible-server virtual-endpoint'] = """
type: group
short-summary: Manage virtual endpoints for a PostgreSQL flexible server.
"""

helps['postgres flexible-server virtual-endpoint create'] = """
type: command
short-summary: Create a new virtual endpoint for a flexible server.
examples:
  - name: >
      Create a virtual endpoint with writer endpoint.
    text: >
      az postgres flexible-server virtual-endpoint create --resource-group testGroup --server-name testserver --name test-virtual-endpoint --endpoint-type ReadWrite --members testReplica1
"""

helps['postgres flexible-server virtual-endpoint list'] = """
type: command
short-summary: List all virtual endpoints for a flexible server.
example:
  - name: List all virtual endpoints for a flexible server.
    text: az postgres flexible-server virtual-endpoint list --resource-group testGroup --server-name testserver
"""

helps['postgres flexible-server virtual-endpoint show'] = """
type: command
short-summary: Get the details of a virtual endpoint.
examples:
  - name: Get the details of a virtual endpoint.
    text: az postgres flexible-server virtual-endpoint show --resource-group testGroup --server-name testserver --name test-virtual-endpoint
"""

helps['postgres flexible-server virtual-endpoint update'] = """
type: command
short-summary: Update a virtual endpoint.
examples:
  - name: Update a virtual endpoint.
    text: az postgres flexible-server virtual-endpoint update --resource-group testGroup --server-name testserver --name test-virtual-endpoint --endpoint-type ReadWrite --members testReplica1
"""

helps['postgres flexible-server virtual-endpoint delete'] = """
type: command
short-summary: Delete a virtual endpoint.
examples:
  - name: Delete a virtual endpoint.
    text: az postgres flexible-server virtual-endpoint delete --resource-group testGroup --server-name testserver --name test-virtual-endpoint
"""

helps['postgres flexible-server migration'] = """
type: group
short-summary: Manage migration workflows for PostgreSQL Flexible Servers.
"""

helps['postgres flexible-server migration create'] = """
type: command
short-summary: Create a new migration workflow for a flexible server.
examples:
  - name: >
      Start a migration workflow on the target server identified by the parameters. The configurations of the migration should be specified in the properties file. The different properties are defined as: \n
      sourceDbServerResourceId: Source server details. \n
      adminCredentials: This parameter lists passwords for admin users for both the source server and the target PostgreSQL flexible server. \n
      targetServerUserName: The default value is the admin user created during the creation of the PostgreSQL target flexible server, and the password provided is used for authentication against this user. \n
      dbsToMigrate: Specify the list of databases that you want to migrate to Flexible Server. \n
      overwriteDBsInTarget: When set to true (default), if the target server happens to have an existing database with the same name as the one you're trying to migrate, the migration service automatically overwrites the database. \n
      Sample migrationConfig.json for PostgreSQLSingleServer shown below. \n
      {
        "properties": {
          "sourceDBServerResourceId": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/test-single-rg/providers/Microsoft.DBforPostgreSQL/servers/pg-single-1",
          "secretParameters": {
            "adminCredentials": {
              "sourceServerPassword": "password",
              "targetServerPassword": "password"
            },
            "sourceServerUserName": "testuser@pg-single-1",
            "targetServerUserName": "fspguser"
          },
          "dBsToMigrate": [
            "postgres"
          ],
          "overwriteDbsInTarget": "true"
        }
      }
    text: >
      az postgres flexible-server migration create --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver \
      --migration-name testmigration --properties "migrationConfig.json"
  - name: >
      Start a migration workflow on the target server identified by the parameters. The configurations of the migration should be specified in the migrationConfig.json file. \
      Use --migration-mode offline for Offline migration.
    text: >
      az postgres flexible-server migration create --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver \
      --migration-name testmigration --properties "migrationConfig.json" --migration-mode offline
  - name: >
      Start a migration workflow on the target server identified by the parameters. The configurations of the migration should be specified in the migrationConfig.json file. \
      Use --migration-mode online for Online(with CDC) migration. Use migration-option Validate for validate only request.
    text: >
      az postgres flexible-server migration create --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver \
      --migration-name testmigration --properties "migrationConfig.json" --migration-mode online --migration-option Validate
  - name: >
      Start a migration workflow on the target server identified by the parameters. The configurations of the migration should be specified in the migrationConfig.json file. \
      Use --migration-option Migrate for Migrate Only request.
    text: >
      az postgres flexible-server migration create --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver \
      --migration-name testmigration --properties "migrationConfig.json" --migration-option Migrate
  - name: >
      To start a migration for other than PostgreSQLSingleServer, soureType and sslMode must be specified in properties file. These properties are defined as: \n
      sourceType: Values can be - OnPremises, AWS_AURORA, AWS_RDS, AzureVM, PostgreSQLSingleServer \n
      sslMode:  SSL modes for migration. SSL mode for PostgreSQLSingleServer is VerifyFull and Prefer/Require for other source types. \n
      Sample migrationConfig.json shown below. \n
      {
        "properties": {
          "sourceDBServerResourceId": "<<hostname or IP address>>:<<port>>@<<username>>",
          "secretParameters": {
            "adminCredentials": {
              "sourceServerPassword": "password",
              "targetServerPassword": "password"
            },
            "sourceServerUserName": "postgres",
            "targetServerUserName": "fspguser"
          },
          "dBsToMigrate": [
            "ticketdb","timedb","inventorydb"
          ],
          "overwriteDbsInTarget": "true",
          "sourceType": "OnPremises",
          "sslMode": "Prefer"
        }
      }
    text: >
      az postgres flexible-server migration create --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver \
        --migration-name testmigration --properties "migrationConfig.json"
  - name: >
      Start a private endpoint enabled migration workflow on the target server by specifying migrationRuntimeResourceId in properties file. This property is defined as: \n
      migrationRuntimeResourceId: The resource ID of the migration runtime server that is responsible for migrating data between source and target server. \n
      Sample migrationConfig.json shown below. \n
      {
        "properties": {
          "sourceDBServerResourceId": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/test-single-rg/providers/Microsoft.DBforPostgreSQL/servers/pg-single-1",
          "migrationRuntimeResourceId": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/testGroup/providers/Microsoft.DBforPostgreSQL/flexibleServers/testsourcemigration",
          "secretParameters": {
            "adminCredentials": {
              "sourceServerPassword": "password",
              "targetServerPassword": "password"
            },
            "sourceServerUserName": "testuser@pg-single-1",
            "targetServerUserName": "fspguser"
          },
          "dBsToMigrate": [
            "postgres"
          ],
          "overwriteDbsInTarget": "true"
        }
      }
    text: >
      az postgres flexible-server migration create --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --migration-name testmigration
      --properties "migrationConfig.json"
"""

helps['postgres flexible-server migration list'] = """
type: command
short-summary: List the migrations of a flexible server.
examples:
  - name: List the currently active migrations of a target flexible server.
    text: az postgres flexible-server migration list --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --filter Active
  - name: List all (Active/Completed) migrations of a target flexible server.
    text: az postgres flexible-server migration list --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --filter All
"""

helps['postgres flexible-server migration show'] = """
type: command
short-summary: Get the details of a specific migration.
examples:
  - name: Get the details of a specific migration of a target flexible server.
    text: az postgres flexible-server migration show --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --migration-name testmigration
"""

helps['postgres flexible-server migration update'] = """
type: command
short-summary: Update a specific migration.
examples:
  - name: Allow the migration workflow to setup logical replication on the source. Note that this command will restart the source server.
    text: az postgres flexible-server migration update --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --migration-name testmigration --setup-replication
  - name: Cut-over the data migration for all the databases involved in the migration. After this is complete, subsequent updates to all databases in the migration will not be migrated to the target. Cutover migration can only be triggered for migration_mode=Online.
    text: az postgres flexible-server migration update --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --migration-name testmigration --cutover
  - name: Cancels the data migration for all the databases involved in the migration. Only 'InProgress' migration can be cancelled
    text: az postgres flexible-server migration update --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --migration-name testmigration --cancel
"""

helps['postgres flexible-server migration check-name-availability'] = """
type: command
short-summary: Checks if the provided migration-name can be used.
examples:
  - name: Check if the migration-name provided is available for your migration workflow.
    text: az postgres flexible-server migration check-name-availability --subscription xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx --resource-group testGroup --name testserver --migration-name xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
"""

helps['postgres flexible-server parameter'] = """
type: group
short-summary: Commands for managing server parameter values for flexible server.
"""

helps['postgres flexible-server parameter list'] = """
type: command
short-summary: List the parameter values for a flexible server.
examples:
  - name: List the parameter values for a flexible server.
    text: az postgres flexible-server parameter list --resource-group testGroup --server-name servername
  - name: List the parameter values for a flexible server in table format.
    text: az postgres flexible-server parameter list --resource-group testGroup --server-name servername --output table
"""

helps['postgres flexible-server parameter set'] = """
type: command
short-summary: Update the parameter of a flexible server.
examples:
  - name: Set a new parameter value.
    text: az postgres flexible-server parameter set --resource-group testGroup --server-name servername --name parameterName --value parameterValue
  - name: Set a parameter value to its default.
    text: az postgres flexible-server parameter set --resource-group testGroup --server-name servername --name parameterName
"""

helps['postgres flexible-server parameter show'] = """
type: command
short-summary: Get the parameter for a flexible server."
examples:
  - name: Get the parameter for a server.
    text: az postgres flexible-server parameter show --resource-group testGroup --server-name servername --name parameterName
"""

helps['postgres flexible-server list-skus'] = """
type: command
short-summary: Lists available sku's in the given region.
example:
  - name: Lists available sku's in the given region.
    text: az postgres flexible-server list-skus -l eastus --output table
"""

helps['postgres flexible-server show-connection-string'] = """
type: command
short-summary: Show the connection strings for a PostgreSQL flexible-server database.
examples:
  - name: Show connection strings for cmd and programming languages.
    text: az postgres flexible-server show-connection-string -s testserver -u username -p password -d databasename
  - name: Show connection strings for cmd and programming languages with PgBouncer enabled.
    text: az postgres flexible-server show-connection-string -s testserver -u username -p password -d databasename --pg-bouncer
"""

helps['postgres flexible-server deploy'] = """
type: group
short-summary: Enable and run GitHub Actions workflow for PostgreSQL server
"""

helps['postgres flexible-server deploy setup'] = """
type: command
short-summary: Create GitHub Actions workflow file for PostgreSQL server.
examples:
  - name: Create GitHub Actions workflow file for PostgreSQL server.
    text: az postgres flexible-server deploy setup -s testserver -g testGroup -u username -p password --sql-file test.sql --repo username/userRepo -d flexibleserverdb --action-name testAction
  - name: Create GitHub Actions workflow file for PostgreSQL server and push it to the remote repository
    text: az postgres flexible-server deploy setup -s testserver -g testGroup -u username -p password --sql-file test.sql --repo username/userRepo -d flexibleserverdb --action-name testAction --branch userBranch --allow-push
"""

helps['postgres flexible-server deploy run'] = """
type: command
short-summary: Run an existing workflow in your github repository
examples:
  - name: Run an existing workflow in your github repository
    text: az postgres flexible-server deploy run --action-name testAction --branch userBranch
"""

helps['postgres flexible-server backup'] = """
type: group
short-summary: Manage flexible server backups.
"""

helps['postgres flexible-server backup create'] = """
type: command
short-summary: Create a new backup for a flexible server.
examples:
  - name: >
      Create a backup.
    text: >
      az postgres flexible-server backup create -g testgroup -n testsvr --backup-name testbackup
"""

helps['postgres flexible-server backup list'] = """
type: command
short-summary: List all the backups for a given server.
examples:
  - name: List all backups for 'testsvr'.
    text: az postgres flexible-server backup list -g testgroup -n testsvr
"""

helps['postgres flexible-server backup show'] = """
type: command
short-summary: Show the details of a specific backup for a given server.
examples:
  - name: Show the details of backup 'testbackup' for 'testsvr'.
    text: az postgres flexible-server backup show -g testgroup -n testsvr --backup-name testbackup
"""

helps['postgres flexible-server backup delete'] = """
type: command
short-summary: Delete a specific backup.
examples:
  - name: Delete a backup.
    text: az postgres flexible-server backup delete -g testgroup -n testsvr --backup-name testbackup
"""

helps['postgres flexible-server long-term-retention'] = """
type: group
short-summary: Manage flexible server long-term-retention backups.
"""

helps['postgres flexible-server long-term-retention pre-check'] = """
type: command
short-summary: Performs all the checks that are needed for the subsequent long-term-retention backup operation to succeed.
examples:
  - name: Precheck if we can perform long-term-retention command on server 'server-name' on resource group 'resource-group-name' with backup name 'backup-name'.
    text: az postgres flexible-server long-term-retention pre-check -g resource-group-name -b backup-name -n server-name
"""

helps['postgres flexible-server long-term-retention start'] = """
type: command
short-summary: Start long-term-retention backup for a flexible server. SAS URL parameter refers to the container SAS URL, inside the storage account, where the backups will be uploaded.
examples:
  - name: Create a backup with name 'backup-name' of server 'server-name' in resource group 'resource-group-name', using container with SAS URL '<sas-url>'.
    text: az postgres flexible-server long-term-retention start -g resource-group-name -b backup-name -n server-name -u <sas-url>
"""

helps['postgres flexible-server long-term-retention show'] = """
type: command
short-summary: Show the details of a specific long-term-retention backup for a given server.
examples:
  - name: Show the details of long-term-retention backup 'testbackup' for 'testsvr'.
    text: az postgres flexible-server long-term-retention show -g resource-group-name -n server-name -b backup-name
"""

helps['postgres flexible-server long-term-retention list'] = """
type: command
short-summary: List all the long-term-retention backups for a given server.
examples:
  - name: List all long-term-retention backups for 'testsvr'.
    text: az postgres flexible-server long-term-retention list -g resource-group-name -n server-name
"""

helps['postgres flexible-server replica'] = """
type: group
short-summary: Manage read replicas.
"""

helps['postgres flexible-server replica create'] = """
type: command
short-summary: Create a read replica for a server.
examples:
  - name: Create a read replica 'testreplicaserver' for 'testserver' with public or private access in the specified zone and location if available.
    text: az postgres flexible-server replica create --replica-name testreplicaserver -g testGroup --source-server testserver --zone 3 --location testLocation
  - name: Create a read replica 'testreplicaserver' with new subnet for 'testserver' with private access.
    text: >
      az postgres flexible-server replica create --replica-name testreplicaserver -g testGroup \\
        --source-server testserver --zone 3 --location testLocation \\
        --vnet newVnet --subnet newSubnet \\
        --address-prefixes 172.0.0.0/16 --subnet-prefixes 172.0.0.0/24 \\
        --private-dns-zone testDNS.postgres.database.azure.com \\
        --tags "key=value"
  - name: >
      Create a read replica 'testreplicaserver' for 'testserver' with public or private access \
      in the specified location if available. Since zone is not passed, it will automatically pick up zone in the \
      replica location which is different from source server, if available, else will pick up zone same as source server \
      in the replica location if available, else will set the zone as None, i.e. No preference
    text: az postgres flexible-server replica create --replica-name testreplicaserver -g testGroup --source-server testserver --location testLocation
  - name: Create a read replica 'testreplicaserver' for 'testserver' with custom --storage-size and --sku.
    text: az postgres flexible-server replica create --replica-name testreplicaserver -g testGroup --source-server testserver --sku-name Standard_D4ds_v5 --storage-size 256
  - name: Create a read replica 'testreplicaserver' for 'testserver', where 'testreplicaserver' is in a different resource group 'newTestGroup'. \
      Here --resource-group is for the read replica's resource group, and --source-server must be passed as resource ID.
    text: >
      az postgres flexible-server replica create --replica-name testreplicaserver -g newTestGroup \
        --source-server /subscriptions/{sourceSubscriptionId}/resourceGroups/{sourceResourceGroup}/providers/Microsoft.DBforPostgreSQL/flexibleServers/{sourceServerName} --location testLocation
"""

helps['postgres flexible-server replica list'] = """
type: command
short-summary: List all read replicas for a given server.
examples:
  - name: List all read replicas for master server 'testserver'.
    text: az postgres flexible-server replica list -g testGroup -n testserver
"""

helps['postgres flexible-server replica promote'] = """
type: command
short-summary: Stop replication of a read replica and promote it to an independent server or as a primary server.
examples:
  - name: Stop replication to 'testreplicaserver' and promote it a standalone read/write server.
    text: az postgres flexible-server replica promote -g testGroup -n testreplicaserver
  - name: Stop replication to 'testreplicaserver' and promote it a standalone read/write server with forced data sync.
    text: az postgres flexible-server replica promote -g testGroup -n testreplicaserver --promote-mode standalone --promote-option forced
  - name: >
      Stop replication to 'testreplicaserver' and promote it to primary server with planned data sync. \
      The replica you are promoting must have the reader virtual endpoint assigned, or you will receive an error on promotion.
    text: az postgres flexible-server replica promote -g testGroup -n testreplicaserver --promote-mode switchover --promote-option planned
"""

helps['postgres flexible-server geo-restore'] = """
type: command
short-summary: Geo-restore a flexible server from backup.
examples:
  - name: >
      Geo-restore public access server 'testserver' to a new server 'testserverNew' in location 'newLocation' with public access.
    text: az postgres flexible-server geo-restore --resource-group testGroup --name testserverNew --source-server testserver --location newLocation
  - name: >
      Geo-restore private access server 'testserver' as a new server 'testserverNew' with new subnet.
      New vnet, subnet, and private dns zone for the restored server will be provisioned. Please refer to 'flexible-server create' command for more private access scenarios.
    text: >
      az postgres flexible-server geo-restore --resource-group testGroup --name testserverNew \\
        --source-server testserver --vnet newVnet --subnet newSubnet \\
        --address-prefixes 172.0.0.0/16 --subnet-prefixes 172.0.0.0/24 \\
        --private-dns-zone testDNS.postgres.database.azure.com --location newLocation
  - name: >
      Geo-restore 'testserver' to current point-in-time as a new server 'testserverNew' in a different subscription / resource group. \\
      Here --resource-group is for the target server's resource group, and --source-server must be passed as resource ID. \\
      This resource ID can be in a subscription different than the subscription used for az account set.
    text: >
      az postgres flexible-server geo-restore --resource-group testGroup --name testserverNew --location newLocation \\
        --source-server /subscriptions/{sourceSubscriptionId}/resourceGroups/{sourceResourceGroup}/providers/Microsoft.DBforPostgreSQL/flexibleServers/{sourceServerName}
"""

helps['postgres flexible-server revive-dropped'] = """
type: command
short-summary: Revive a dropped flexible server from backup.
examples:
  - name: >
      Revive dropped public access server 'testserver' to a new server 'testserverNew' in location 'newLocation' with public access.
    text: az postgres flexible-server revive-dropped --resource-group testGroup --name testserverNew --source-server /subscriptions/{SubId}/resourceGroups/{testGroup}/providers/Microsoft.DBforPostgreSQL/flexibleServers/testserver --location newLocation
  - name: >
      Revive dropped public access server 'testserver' with data encryption enabled as a new server 'testserverNew' with data encryption.
    text: >
      az postgres flexible-server revive-dropped -l testLocation --resource-group testGroup --name testserverNew \\
        --source-server testserver --key newKeyIdentifier --identity newIdentity
"""

helps['postgres flexible-server upgrade'] = """
type: command
short-summary: Upgrade the major version of a flexible server.
examples:
  - name: Upgrade server 'testsvr' to PostgreSQL major version 17.
    text: az postgres flexible-server upgrade -g testgroup -n testsvr -v 17
"""

helps['postgres flexible-server identity'] = """
type: group
short-summary: Manage server user assigned identities.
"""

helps['postgres flexible-server identity update'] = """
type: command
short-summary: Update to enable or disable system assigned managed identity on the server.
examples:
  - name: Enable system assigned managed identity on the server.
    text: az postgres flexible-server identity update -g testgroup -s testsvr --system-assigned Enabled
"""

helps['postgres flexible-server identity assign'] = """
type: command
short-summary: Add user assigned managed identities to the server.
examples:
  - name: Add identities 'test-identity' and 'test-identity-2' to server 'testsvr'.
    text: az postgres flexible-server identity assign -g testgroup -s testsvr --identity test-identity test-identity-2
"""

helps['postgres flexible-server identity remove'] = """
type: command
short-summary: Remove user assigned managed identites from the server.
examples:
  - name: Remove identity 'test-identity' from server 'testsvr'.
    text: az postgres flexible-server identity remove -g testgroup -s testsvr --identity test-identity
"""

helps['postgres flexible-server identity show'] = """
type: command
short-summary: Get an user assigned managed identity from the server.
examples:
  - name: Get identity 'test-identity' from server 'testsvr'.
    text: az postgres flexible-server identity show -g testgroup -s testsvr --identity test-identity
"""

helps['postgres flexible-server identity list'] = """
type: command
short-summary: List all user assigned managed identities from the server.
examples:
  - name: List all identities from server 'testsvr'.
    text: az postgres flexible-server identity list -g testgroup -s testsvr
"""

helps['postgres flexible-server microsoft-entra-admin'] = """
type: group
short-summary: Manage server Microsoft Entra administrators.
"""

helps['postgres flexible-server microsoft-entra-admin create'] = """
type: command
short-summary: Create a Microsoft Entra administrator.
examples:
  - name: Create Microsoft Entra administrator with user 'john@contoso.com', administrator ID '00000000-0000-0000-0000-000000000000' and type User.
    text: az postgres flexible-server microsoft-entra-admin create -g testgroup -s testsvr -u john@contoso.com -i 00000000-0000-0000-0000-000000000000 -t User
"""

helps['postgres flexible-server microsoft-entra-admin delete'] = """
type: command
short-summary: Delete a Microsoft Entra administrator.
examples:
  - name: Delete Microsoft Entra administrator with ID '00000000-0000-0000-0000-000000000000'.
    text: az postgres flexible-server microsoft-entra-admin delete -g testgroup -s testsvr -i 00000000-0000-0000-0000-000000000000
"""

helps['postgres flexible-server microsoft-entra-admin list'] = """
type: command
short-summary: List all Microsoft Entra administrators.
examples:
  - name: List Microsoft Entra administrators.
    text: az postgres flexible-server microsoft-entra-admin list -g testgroup -s testsvr
"""

helps['postgres flexible-server microsoft-entra-admin show'] = """
type: command
short-summary: Get a Microsoft Entra administrator.
examples:
  - name: Get Microsoft Entra administrator with ID '00000000-0000-0000-0000-000000000000'.
    text: az postgres flexible-server microsoft-entra-admin show -g testgroup -s testsvr -i 00000000-0000-0000-0000-000000000000
"""

helps['postgres flexible-server microsoft-entra-admin wait'] = """
type: command
short-summary: Wait for a Microsoft Entra administrator to satisfy certain conditions.
examples:
  - name: Wait until a Microsoft Entra administrator exists.
    text: az postgres flexible-server microsoft-entra-admin wait -g testgroup -s testsvr -i 00000000-0000-0000-0000-000000000000 --exists
  - name: Wait for a Microsoft Entra administrator to be deleted.
    text: az postgres flexible-server microsoft-entra-admin wait -g testgroup -s testsvr -i 00000000-0000-0000-0000-000000000000 --deleted
"""

helps['postgres flexible-server advanced-threat-protection-setting'] = """
type: group
short-summary: Manage advanced threat protection setting for a PostgreSQL flexible server.
"""

helps['postgres flexible-server advanced-threat-protection-setting update'] = """
type: command
short-summary: Updates advanced threat protection setting state for a flexible server.
examples:
  - name: >
      Enable advanced threat protection setting for a PostgreSQL flexible server.
    text: >
      az postgres flexible-server advanced-threat-protection-setting update --resource-group testGroup --server-name testserver --state Enabled
  - name: >
      Disable advanced threat protection setting for a PostgreSQL flexible server.
    text: >
      az postgres flexible-server advanced-threat-protection-setting update --resource-group testGroup --server-name testserver --state Disabled
"""

helps['postgres flexible-server advanced-threat-protection-setting show'] = """
type: command
short-summary: Get advanced threat protection settings for a PostgreSL flexible server.
examples:
  - name: Get the details of advanced threat protection setting for a flexible server.
    text: az postgres flexible-server advanced-threat-protection-setting show --resource-group testGroup --server-name testserver
  - name: Get the details of advanced threat protection setting for a flexible server in a different subscription.
    text: az postgres flexible-server advanced-threat-protection-setting show --subscription testSubscription --resource-group testGroup --server-name testserver
  - name: Get the details of advanced threat protection setting for a flexible server using --ids parameter.
    text: az postgres flexible-server advanced-threat-protection-setting show --ids /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/testGroup/providers/Microsoft.DBforPostgreSQL/flexibleServers/testServer
"""

helps['postgres flexible-server server-logs'] = """
type: group
short-summary: Manage server logs (log files) for a PostgreSQL flexible server.
"""

helps['postgres flexible-server server-logs download'] = """
type: command
short-summary: Download log files for a PostgreSQL flexible server.
examples:
  - name: >
      Downloads log files f1 and f2 to the current directory from the server 'testsvr'. Please note that f1 and f2 should match the log file name including the foldername, for instance serverlogs/f1.log
    text: >
      az postgres flexible-server server-logs download -g testgroup -s testsvr -n serverlogs/f1.log serverlogs/f2.log
"""

helps['postgres flexible-server server-logs list'] = """
type: command
short-summary: List log files for a PostgreSQL flexible server.
examples:
  - name: List log files for 'testsvr' modified in the last 72 hours (default value).
    text: az postgres flexible-server server-logs list -g testgroup -s testsvr
  - name: List log files for 'testsvr' modified in the last 10 hours.
    text: az postgres flexible-server server-logs list -g testgroup -s testsvr --file-last-written 10
  - name: List log files for 'testsvr' less than 30Kb in size.
    text: az postgres flexible-server server-logs list -g testgroup -s testsvr --max-file-size 30
  - name: List log files for 'testsvr' containing name 'serverlogs'.
    text: az postgres flexible-server server-logs list -g testgroup -s testsvr --subscription testSubscription --filename-contains serverlogs
"""

helps['postgres flexible-server private-endpoint-connection'] = """
type: group
short-summary: Manage PostgreSQL flexible server private endpoint connections.
"""

helps['postgres flexible-server private-endpoint-connection list'] = """
type: command
short-summary: List all private endpoint connections associated with a PostgreSQL flexible server.
examples:
  - name: List all private endpoint connections associated with a PostgreSQL flexible server.
    text: az postgres flexible-server private-endpoint-connection list -g testgroup -s testsvr
"""

helps['postgres flexible-server private-endpoint-connection show'] = """
type: command
short-summary: Show details of a private endpoint connection associated with a PostgreSQL flexible server.
examples:
  - name: Show details of a private endpoint connection associated with a PostgreSQL flexible server.
    text: az postgres flexible-server private-endpoint-connection show -g testgroup -s testsvr -n pec-connection.40e3d3a8-7d8f-41eb-8462-1cd05bc3e33b
"""

helps['postgres flexible-server private-endpoint-connection approve'] = """
type: command
short-summary: Approve the specified private endpoint connection associated with a PostgreSQL flexible server.
examples:
  - name: Approve a private endpoint connection associated with a PostgreSQL flexible server.
    text: >
      az postgres flexible-server private-endpoint-connection approve -g testgroup -s testsvr \
        -n pec-connection.40e3d3a8-7d8f-41eb-8462-1cd05bc3e33b \
        --description "Approve connection"
"""

helps['postgres flexible-server private-endpoint-connection reject'] = """
type: command
short-summary: Reject the specified private endpoint connection associated with a PostgreSQL flexible server.
examples:
  - name: Reject a private endpoint connection associated with a PostgreSQL flexible server.
    text: >
      az postgres flexible-server private-endpoint-connection reject -g testgroup -s testsvr \
        -n pec-connection.40e3d3a8-7d8f-41eb-8462-1cd05bc3e33b \
        --description "Reject connection"
"""

helps['postgres flexible-server private-endpoint-connection delete'] = """
type: command
short-summary: Delete the specified private endpoint connection associated with a PostgreSQL flexible server.
examples:
  - name: Delete a private endpoint connection associated with a PostgreSQL flexible server.
    text: >
      az postgres flexible-server private-endpoint-connection delete -g testgroup -s testsvr \
        --id /subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/testgroup/providers/Microsoft.DBforPostgreSQL/flexibleServers/testsvr/privateEndpointConnections/pec-connection.40e3d3a8-7d8f-41eb-8462-1cd05bc3e33b
"""

helps['postgres flexible-server private-link-resource'] = """
type: group
short-summary: Get Private link resource for a PostgreSQL flexible server.
"""

helps['postgres flexible-server private-link-resource list'] = """
type: command
short-summary: List private link resources associated with a PostgreSQL flexible server.
examples:
  - name: List private link resources associated with a PostgreSQL flexible server.
    text: az postgres flexible-server private-link-resource list -g testgroup -s testsvr
"""

helps['postgres flexible-server private-link-resource show'] = """
type: command
short-summary: Get private link  resource for a PostgreSQL flexible server.
examples:
  - name: Get the private link  resource for a flexible server.
    text: az postgres flexible-server private-link-resource show --resource-group testGroup --server-name testserver
  - name: Get the private link  resource for a flexible server in a different subscription.
    text: az postgres flexible-server private-link-resource show --subscription testSubscription --resource-group testGroup --server-name testserver
  - name: Get the private link resource for a flexible server using --ids parameter.
    text: az postgres flexible-server private-link-resource show --ids /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/testGroup/providers/Microsoft.DBforPostgreSQL/flexibleServers/testServer
"""

helps['postgres flexible-server fabric-mirroring'] = """
type: group
short-summary: Bring your PostgreSQL data into Microsoft Fabric. Mirroring allows you to create a replica of your data in OneLake which can be used for all your analytical needs.
"""

helps['postgres flexible-server fabric-mirroring start'] = """
type: command
short-summary: Enable bringing your PostgreSQL data into Microsoft Fabric.
examples:
  - name: Enable bringing your PostgreSQL data into Microsoft Fabric.
    text: az postgres flexible-server fabric-mirroring start -g testgroup -s testsvr --database-names testdb
"""

helps['postgres flexible-server fabric-mirroring stop'] = """
type: command
short-summary: Stop bringing your PostgreSQL data into Microsoft Fabric.
examples:
  - name: Stop bringing your PostgreSQL data into Microsoft Fabric.
    text: az postgres flexible-server fabric-mirroring stop -g testgroup -s testsvr
"""

helps['postgres flexible-server fabric-mirroring update-databases'] = """
type: command
short-summary: Update allowed mirrored databases.
examples:
  - name: Update allowed mirrored databases.
    text: az postgres flexible-server fabric-mirroring update-databases -g testgroup -s testsvr --database-names testdb2 testdb3
"""

helps['postgres flexible-server index-tuning'] = """
type: group
short-summary: Index tuning analyzes read queries captured in Query Store and recommends index changes to optimize these queries.
"""

helps['postgres flexible-server index-tuning update'] = """
type: command
short-summary: Update index tuning to be enabled/disabled for a PostgreSQL flexible server.
examples:
  - name: Update index tuning to be enabled/disabled for a PostgreSQL flexible server.
    text: az postgres flexible-server index-tuning update -g testgroup -s testsvr --enabled True
"""

helps['postgres flexible-server index-tuning show'] = """
type: command
short-summary: Show state of index tuning for a PostgreSQL flexible server.
examples:
  - name: Show state of index tuning for a PostgreSQL flexible server.
    text: az postgres flexible-server index-tuning show -g testgroup -s testsvr
"""

helps['postgres flexible-server index-tuning list-settings'] = """
type: command
short-summary: Get tuning settings associated for a PostgreSQL flexible server.
examples:
  - name: Get tuning settings for a PostgreSQL flexible server.
    text: az postgres flexible-server index-tuning list-settings -g testgroup -s testsvr
"""

helps['postgres flexible-server index-tuning show-settings'] = """
type: command
short-summary: Get a tuning setting for a PostgreSQL flexible server.
examples:
  - name: Get a tuning setting for a PostgreSQL flexible server.
    text: az postgres flexible-server index-tuning show-settings -g testgroup -s testsvr --name setting-name
"""

helps['postgres flexible-server index-tuning set-settings'] = """
type: command
short-summary: Update a tuning setting for a PostgreSQL flexible server.
examples:
  - name: Update a tuning setting for a PostgreSQL flexible server.
    text: az postgres flexible-server index-tuning set-settings -g testgroup -s testsvr --name setting-name --value setting-value
"""

helps['postgres flexible-server index-tuning list-recommendations'] = """
type: command
short-summary: Get available tuning index recommendations associated with a PostgreSQL flexible server.
examples:
  - name: Get tuning index recommendations for a PostgreSQL flexible server. Filter by selected type.
    text: az postgres flexible-server index-tuning list-recommendations -g testgroup -s testsvr --recommendation-type CreateIndex
"""



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def budget_output(result):
    result['amount'] = str(result['amount'])
    if 'currentSpend' in result:
        result['currentSpend']['amount'] = str(result['currentSpend'].get('amount', None))
    if 'notifications' in result:
        for key in result['notifications']:
            value = result['notifications'][key]
            value['threshold'] = str(value.get('threshold', None))
    return result


def transform_budget_show_output(result):
    return budget_output(result)


def transform_budget_create_update_output(result):
    return budget_output(result)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "eventhubs namespace application-group create",
)
class Create(AAZCommand):
    """Create an ApplicationGroup for a Namespace.
    """

    _aaz_info = {
        "version": "2023-01-01-preview",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.eventhub/namespaces/{}/applicationgroups/{}", "2023-01-01-preview"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.application_group_name = AAZStrArg(
            options=["-n", "--name", "--application-group-name"],
            help="The Application Group name ",
            required=True,
            fmt=AAZStrArgFormat(
                max_length=256,
                min_length=1,
            ),
        )
        _args_schema.namespace_name = AAZStrArg(
            options=["--namespace-name"],
            help="The Namespace name",
            required=True,
            fmt=AAZStrArgFormat(
                max_length=50,
                min_length=6,
            ),
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.client_app_group_identifier = AAZStrArg(
            options=["--client-app-group-id", "--client-app-group-identifier"],
            arg_group="Properties",
            help="The Unique identifier for application group.Supports SAS(SASKeyName=KeyName) or AAD(AADAppID=Guid)",
        )
        _args_schema.is_enabled = AAZBoolArg(
            options=["--is-enabled"],
            arg_group="Properties",
            help="Determines if Application Group is allowed to create connection with namespace or not. Once the isEnabled is set to false, all the existing connections of application group gets dropped and no new connections will be allowed",
        )
        _args_schema.policies = AAZListArg(
            options=["--policies"],
            arg_group="Properties",
            help="List of group policies that define the behavior of application group. The policies can support resource governance scenarios such as limiting ingress or egress traffic.",
        )

        policies = cls._args_schema.policies
        policies.Element = AAZObjectArg()

        _element = cls._args_schema.policies.Element
        _element.throttling_policy = AAZObjectArg(
            options=["throttling-policy"],
        )
        _element.name = AAZStrArg(
            options=["name"],
            help="The Name of this policy",
            required=True,
        )

        throttling_policy = cls._args_schema.policies.Element.throttling_policy
        throttling_policy.metric_id = AAZStrArg(
            options=["metric-id"],
            help="Metric Id on which the throttle limit should be set, MetricId can be discovered by hovering over Metric in the Metrics section of Event Hub Namespace inside Azure Portal",
            required=True,
            enum={"IncomingBytes": "IncomingBytes", "IncomingMessages": "IncomingMessages", "OutgoingBytes": "OutgoingBytes", "OutgoingMessages": "OutgoingMessages"},
        )
        throttling_policy.rate_limit_threshold = AAZIntArg(
            options=["rate-limit-threshold"],
            help="The Threshold limit above which the application group will be throttled.Rate limit is always per second.",
            required=True,
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        self.ApplicationGroupCreateOrUpdateApplicationGroup(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class ApplicationGroupCreateOrUpdateApplicationGroup(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.EventHub/namespaces/{namespaceName}/applicationGroups/{applicationGroupName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "applicationGroupName", self.ctx.args.application_group_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "namespaceName", self.ctx.args.namespace_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2023-01-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("clientAppGroupIdentifier", AAZStrType, ".client_app_group_identifier", typ_kwargs={"flags": {"required": True}})
                properties.set_prop("isEnabled", AAZBoolType, ".is_enabled")
                properties.set_prop("policies", AAZListType, ".policies")

            policies = _builder.get(".properties.policies")
            if policies is not None:
                policies.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.policies[]")
            if _elements is not None:
                _elements.set_prop("name", AAZStrType, ".name", typ_kwargs={"flags": {"required": True}})
                _elements.set_const("type", "ThrottlingPolicy", AAZStrType, ".throttling_policy", typ_kwargs={"flags": {"required": True}})
                _elements.discriminate_by("type", "ThrottlingPolicy")

            disc_throttling_policy = _builder.get(".properties.policies[]{type:ThrottlingPolicy}")
            if disc_throttling_policy is not None:
                disc_throttling_policy.set_prop("metricId", AAZStrType, ".throttling_policy.metric_id", typ_kwargs={"flags": {"required": True}})
                disc_throttling_policy.set_prop("rateLimitThreshold", AAZIntType, ".throttling_policy.rate_limit_threshold", typ_kwargs={"flags": {"required": True}})

            return self.serialize_content(_content_value)

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.location = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200.system_data = AAZObjectType(
                serialized_name="systemData",
                flags={"read_only": True},
            )
            _schema_on_200.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.properties
            properties.client_app_group_identifier = AAZStrType(
                serialized_name="clientAppGroupIdentifier",
                flags={"required": True},
            )
            properties.is_enabled = AAZBoolType(
                serialized_name="isEnabled",
            )
            properties.policies = AAZListType()

            policies = cls._schema_on_200.properties.policies
            policies.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.policies.Element
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.type = AAZStrType(
                flags={"required": True},
            )

            disc_throttling_policy = cls._schema_on_200.properties.policies.Element.discriminate_by("type", "ThrottlingPolicy")
            disc_throttling_policy.metric_id = AAZStrType(
                serialized_name="metricId",
                flags={"required": True},
            )
            disc_throttling_policy.rate_limit_threshold = AAZIntType(
                serialized_name="rateLimitThreshold",
                flags={"required": True},
            )

            system_data = cls._schema_on_200.system_data
            system_data.created_at = AAZStrType(
                serialized_name="createdAt",
            )
            system_data.created_by = AAZStrType(
                serialized_name="createdBy",
            )
            system_data.created_by_type = AAZStrType(
                serialized_name="createdByType",
            )
            system_data.last_modified_at = AAZStrType(
                serialized_name="lastModifiedAt",
            )
            system_data.last_modified_by = AAZStrType(
                serialized_name="lastModifiedBy",
            )
            system_data.last_modified_by_type = AAZStrType(
                serialized_name="lastModifiedByType",
            )

            return cls._schema_on_200


class _CreateHelper:
    """Helper class for Create"""


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long, too-few-public-methods

import abc
import base64
import json
import os.path
import re
import math

from knack.log import get_logger

from ._command_ctx import AAZCommandCtx
from ._field_type import AAZSimpleType
from ._field_value import AAZUndefined, AAZSimpleValue, AAZDict, AAZList, AAZObject
from .exceptions import AAZInvalidArgValueError

logger = get_logger(__name__)


class AAZBaseArgFormat:

    @abc.abstractmethod
    def __call__(self, ctx: AAZCommandCtx, value):
        raise NotImplementedError()


class AAZStrArgFormat(AAZBaseArgFormat):

    def __init__(self, pattern=None, max_length=None, min_length=None):
        self._pattern = pattern
        self._max_length = max_length
        self._min_length = min_length
        if self._pattern is not None:
            self._compiled_pattern = re.compile(pattern)
        else:
            self._compiled_pattern = None

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        if self._min_length is not None and len(data) < self._min_length:
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' length is less than {self._min_length} ")
        if self._max_length is not None and len(data) > self._max_length:
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' length is greater than {self._max_length}")
        if self._pattern is not None and not self._compiled_pattern.fullmatch(data):
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' does not fully match regular expression pattern '{self._pattern}'")

        return value


class AAZDurationFormat(AAZBaseArgFormat):

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        from msrest.serialization import Serializer
        from isodate.isoerror import ISO8601Error

        try:
            data = Serializer.serialize_duration(data.upper())
        except ISO8601Error:
            try:
                # parse '##DT##H##M##S
                data = Serializer.serialize_duration(f'P0M{data.upper()}')
            except ISO8601Error:
                try:
                    # parse '##H##M##S'
                    data = Serializer.serialize_duration(f'PT{data.upper()}')
                except ISO8601Error:
                    raise AAZInvalidArgValueError(
                        f"Invalid format: '{data}' should be of the form "
                        f"'##dT##h##m##s', '##h##m##s' or ISO8601 duration"
                    )
        value._data = data
        return value


class AAZDateFormat(AAZBaseArgFormat):

    def __init__(self):
        self.help_string = 'Format: date (yyyy-mm-dd)'

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        import dateutil.parser
        import dateutil.tz

        dt_val = None
        try:
            dt_val = dateutil.parser.parse(data)
        except ValueError:
            pass

        if not dt_val:
            raise AAZInvalidArgValueError(
                f"Unable to parse: '{data}'. Expected format: {self.help_string}"
            )

        if any([dt_val.hour, dt_val.minute, dt_val.second, dt_val.microsecond]):
            logger.warning('Time info will be ignored in %s.', data)

        if dt_val.tzinfo:
            logger.warning('Timezone info will be ignored in %s.', data)

        data = "{:04}-{:02}-{:02}".format(dt_val.year, dt_val.month, dt_val.day)
        value._data = data
        return value


class AAZTimeFormat(AAZBaseArgFormat):

    def __init__(self):
        self.help_string = 'Format: time (hh:mm:ss.xxxxxx)'

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        import dateutil.parser
        import dateutil.tz

        dt_val = None
        try:
            dt_val = dateutil.parser.parse(data)
        except ValueError:
            pass

        if not dt_val:
            raise AAZInvalidArgValueError(
                f"Unable to parse: '{data}'. Expected format: {self.help_string}"
            )

        if any([dt_val.day, dt_val.month, dt_val.year]):
            logger.warning('Date info will be ignored in %s.', data)

        if dt_val.tzinfo:
            logger.warning('Timezone info will be ignored in %s.', data)

        data = "{:02}:{:02}:{:02}".format(dt_val.hour, dt_val.minute, dt_val.second)
        if dt_val.microsecond:
            microseconds = str(dt_val.microsecond).rjust(6, '0').rstrip('0').ljust(3, '0')
            data += '.' + microseconds

        value._data = data
        return value


class AAZDateTimeFormat(AAZBaseArgFormat):

    def __init__(self, protocol="iso"):
        self.protocol = protocol
        self.help_string = 'Format: date (yyyy-mm-dd) time (hh:mm:ss.xxxxxx) timezone (+/-hh:mm)'

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        import dateutil.parser
        import dateutil.tz
        from msrest.serialization import Serializer

        dt_val = None
        try:
            dt_val = dateutil.parser.parse(data)
        except ValueError:
            pass

        if not dt_val:
            raise AAZInvalidArgValueError(
                f"Unable to parse: '{data}'. Expected format: {self.help_string}"
            )

        if not dt_val.tzinfo:
            dt_val = dt_val.replace(tzinfo=dateutil.tz.tzlocal())

        if self.protocol == "iso":
            data = Serializer.serialize_iso(dt_val)
        elif self.protocol == "rfc":
            data = Serializer.serialize_rfc(dt_val)
        else:
            raise NotImplementedError()

        value._data = data
        return value


class AAZUuidFormat(AAZBaseArgFormat):

    _uuid_pattern = re.compile(r'^[{(]?[0-9a-fA-F]{8}([-]?[0-9a-fA-F]{4}){3}[-]?[0-9a-fA-F]{12}[)}]?$')

    def __init__(self, case=None, braces_removed=True, hyphens_filled=True):
        """
        :param case: 'upper' to format data into upper case, 'lower' to format data into lower case
        """
        self.case = case
        self.braces_removed = braces_removed
        self.hyphens_filled = hyphens_filled

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        if not self._uuid_pattern.fullmatch(data):
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' is not a valid GUID or UUID"
            )

        if '-' in data and data.count('-') != 4:
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' is not a valid GUID or UUID"
            )

        if data.startswith('{') or data.endswith('}'):
            # remove braces
            if not (data.startswith('{') and data.endswith('}')):
                raise AAZInvalidArgValueError(
                    f"Invalid format: '{data}' is not a valid GUID or UUID"
                )
            if self.braces_removed:
                data = data[1:-1]

        elif data.startswith('(') or data.endswith(')'):
            # remove parentheses
            if not (data.startswith('(') and data.endswith(')')):
                raise AAZInvalidArgValueError(
                    f"Invalid format: '{data}' is not a valid GUID or UUID"
                )
            if self.braces_removed:
                data = data[1:-1]

        if '-' not in data and self.hyphens_filled:
            # add '-' in data
            if data[0] in ('{', '('):
                data = f'{data[:9]}-{data[9:13]}-{data[13:17]}-{data[17:21]}-{data[21:]}'
            else:
                data = f'{data[:8]}-{data[8:12]}-{data[12:16]}-{data[16:20]}-{data[20:]}'

        if self.case == 'upper':
            data = data.upper()
        elif self.case == 'lower':
            data = data.lower()

        value._data = data
        return value


class AAZIntArgFormat(AAZBaseArgFormat):

    def __init__(self, multiple_of=None, maximum=None, minimum=None):
        self._multiple_of = multiple_of
        self._maximum = maximum
        self._minimum = minimum

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, int)
        if self._multiple_of is not None and data % self._multiple_of != 0:
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' is not multiple of {self._multiple_of}")
        if self._minimum is not None and data < self._minimum:
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' is less than {self._minimum}")
        if self._maximum is not None and data > self._maximum:
            raise AAZInvalidArgValueError(
                f"Invalid format: '{data}' is greater than {self._maximum}")
        return value


class AAZFloatArgFormat(AAZBaseArgFormat):

    def __init__(self, multiple_of=None, maximum=None, exclusive_maximum=None, minimum=None, exclusive_minimum=None,
                 rel_tol=1e-09):
        self._multiple_of = multiple_of
        self._maximum = maximum
        self._minimum = minimum
        self._exclusive_maximum = exclusive_maximum
        self._exclusive_minimum = exclusive_minimum
        self._rel_tol = rel_tol

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, float)

        if self._multiple_of is not None:
            q = int(data / self._multiple_of)
            if not math.isclose(data, q * self._multiple_of, rel_tol=self._rel_tol) and \
                    not math.isclose(data, (q + 1) * self._multiple_of, rel_tol=self._rel_tol):
                raise AAZInvalidArgValueError(
                    f"Invalid format: '{data}' is not multiple of {self._multiple_of}")

        if self._minimum is not None:
            if math.isclose(data, self._minimum, rel_tol=self._rel_tol):
                data = self._minimum

            if self._exclusive_minimum:
                if data <= self._minimum:
                    raise AAZInvalidArgValueError(
                        f"Invalid format: '{data}' is not greater than {self._minimum}")
            elif data < self._minimum:
                raise AAZInvalidArgValueError(
                    f"Invalid format: '{data}' is less than {self._minimum}")

        if self._maximum is not None:
            if math.isclose(data, self._maximum, rel_tol=self._rel_tol):
                data = self._maximum

            if self._exclusive_maximum:
                if data >= self._maximum:
                    raise AAZInvalidArgValueError(
                        f"Invalid format: '{data}' is not less than {self._maximum}")
            elif data > self._maximum:
                raise AAZInvalidArgValueError(
                    f"Invalid format: '{data}' is greater than {self._maximum}")

        value._data = data
        return value


class AAZBoolArgFormat(AAZBaseArgFormat):

    def __init__(self, reverse=False):
        self._reverse = reverse  # reverse it's value

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, bool)
        if self._reverse:
            data = not data

        value._data = data
        return value


class AAZObjectArgFormat(AAZBaseArgFormat):

    def __init__(self, max_properties=None, min_properties=None):
        self._max_properties = max_properties
        self._min_properties = min_properties

    def __call__(self, ctx, value):
        assert isinstance(value, AAZObject)
        data = value._data
        if data == AAZUndefined or data is None:
            return value

        assert isinstance(data, dict)

        # Because _fields of object schema is OrderedDict
        # aaz-dev should register arguments in order of their interdependence.
        for field_name, field_schema in value._schema._fields.items():
            if field_schema._options:
                option_name = field_schema._options[0]
            else:
                option_name = field_name
            if field_schema._fmt:
                if field_name in data:
                    try:
                        value[field_name] = field_schema._fmt(ctx, value[field_name])
                    except AAZInvalidArgValueError as err:
                        err.indexes = [option_name, *err.indexes]
                        raise err from err
                else:
                    # build AAZUndefined value and fmt it.
                    v = field_schema._ValueCls(field_schema, AAZUndefined)
                    try:
                        v = field_schema._fmt(ctx, v)
                    except AAZInvalidArgValueError as err:
                        err.indexes = [option_name, *err.indexes]
                        raise err from err
                    if v._data != AAZUndefined:
                        # when result has real data, assign it
                        value[field_name] = v
            if field_schema._required and field_name not in data:
                raise AAZInvalidArgValueError(
                    f"Missing required field: {option_name}")

        if value._is_patch:
            return value

        if self._min_properties and len(data) < self._min_properties:
            raise AAZInvalidArgValueError(
                f"Invalid format: object length is less than {self._min_properties}")

        if self._max_properties and len(data) > self._max_properties:
            raise AAZInvalidArgValueError(
                f"Invalid format: object length is greater than {self._max_properties}")

        return value


class AAZDictArgFormat(AAZBaseArgFormat):

    def __init__(self, max_properties=None, min_properties=None):
        self._max_properties = max_properties
        self._min_properties = min_properties

    def __call__(self, ctx, value):
        assert isinstance(value, AAZDict)
        data = value._data
        if data == AAZUndefined or data is None:
            return value

        assert isinstance(data, dict)
        element_schema = value._schema.Element
        element_fmt = element_schema._fmt
        if element_fmt:
            for key in data:
                try:
                    value[key] = element_fmt(ctx, value[key])
                except AAZInvalidArgValueError as err:
                    err.indexes = [f'[{key}]', *err.indexes]
                    raise err from err

        if value._is_patch:
            return value

        if self._min_properties and len(value) < self._min_properties:
            raise AAZInvalidArgValueError(
                f"Invalid format: dict length is less than {self._min_properties}")

        if self._max_properties and len(value) > self._max_properties:
            raise AAZInvalidArgValueError(
                f"Invalid format: dict length is greater than {self._max_properties}")

        return value


# Warning: This type should not be used any more, the new aaz-dev-tools only use AAZDictArgFormat
class AAZFreeFormDictArgFormat(AAZDictArgFormat):
    pass


class AAZListArgFormat(AAZBaseArgFormat):

    def __init__(self, unique=None, max_length=None, min_length=None):
        self._unique = unique
        self._max_length = max_length
        self._min_length = min_length

    def __call__(self, ctx, value):
        assert isinstance(value, AAZList)
        data = value._data
        if data == AAZUndefined or data is None:
            return value

        assert isinstance(data, dict)
        element_schema = value._schema.Element
        element_fmt = element_schema._fmt
        if element_fmt:
            for idx in data:
                try:
                    value[idx] = element_fmt(ctx, value[idx])
                except AAZInvalidArgValueError as err:
                    err.indexes = [f'[{idx}]', *err.indexes]
                    raise err from err

        if value._is_patch:
            return value

        if self._unique and isinstance(element_schema, AAZSimpleType):
            unique_elements = set()
            for idx in data:
                e = value[idx]
                if e._is_patch:
                    continue
                if e._data in unique_elements:
                    raise AAZInvalidArgValueError(
                        f"Invalid format: '{e._data}' duplicate in list.",
                        indexes=[f'[{idx}]']
                    )
                unique_elements.add(e._data)

        if self._min_length is not None and len(value) < self._min_length:
            raise AAZInvalidArgValueError(
                f"Invalid format: list length is less than {self._min_length} ")

        if self._max_length is not None and len(value) > self._max_length:
            raise AAZInvalidArgValueError(
                f"Invalid format: list length is greater than {self._max_length}")

        return value


class AAZResourceLocationArgFormat(AAZBaseArgFormat):

    def __init__(self, resource_group_arg=None):
        self._resource_group_arg = resource_group_arg

    def __call__(self, ctx, value):
        from ._command_ctx import get_subscription_locations
        assert isinstance(value, AAZSimpleValue)

        data = value._data
        if data is None or value._is_patch:
            return value

        if data == AAZUndefined:
            data = self.get_location_from_resource_group(ctx)

        if data == AAZUndefined:
            return value

        assert isinstance(data, str)
        if ' ' in data:
            for location in get_subscription_locations(ctx):
                if location.display_name.lower() == data.lower():
                    # if display name is provided, attempt to convert to short form name
                    data = location.name
                    break

        value._data = data
        return value

    def get_location_from_resource_group(self, ctx):
        from ._command_ctx import get_resource_group_location

        if not self._resource_group_arg:
            return AAZUndefined

        rg_name = ctx.args[self._resource_group_arg].to_serialized_data()
        if rg_name == AAZUndefined:
            return AAZUndefined

        location = get_resource_group_location(ctx, rg_name)
        if location != AAZUndefined:
            logger.debug("using location '%s' from resource group '%s'", location, rg_name)

        return location


class AAZResourceIdArgFormat(AAZBaseArgFormat):

    class _Template:
        """
        Example template:
            /subscriptions/{subscription}/resourceGroups/{resource_group}/providers/Microsoft.Network/virtualNetworks/{vnet_name}/subnets/{}

        A template string is consisted of several segments joined by '/'.
        Placeholder segment is wrapped by '{}'.

        """

        def __init__(self, template):
            self._template = template
            self._segments = template.split('/')
            blank_placeholders = 0
            for segment in self._segments:
                if segment.startswith('{') != segment.endswith('}'):
                    raise ValueError(f"Invalid segment: {segment}")
                if segment == '{}':
                    blank_placeholders += 1

            # when there's only one '{}' in placeholder, user can only pass the value of that segment.
            self.support_name_segment_only = blank_placeholders == 1

        def __call__(self, ctx, data):
            data_segments = data.split('/')
            if len(data_segments) != len(self._segments) and len(data_segments) == 1 and self.support_name_segment_only:
                # build resource_id by template
                data_segments = []
                for segment in self._segments:
                    if segment == '{}':
                        data_segments.append(data)
                    elif segment.startswith('{'):
                        # use argument value
                        arg_name = segment[1:-1]
                        if arg_name == 'subscription':
                            arg_data = ctx.subscription_id
                        else:
                            arg_data = ctx.args[arg_name].to_serialized_data()

                        if arg_data == AAZUndefined or arg_data is None:
                            raise AAZInvalidArgValueError(
                                f"Invalid format: Failed to construct resource id: argument '{arg_name}' is not defined")
                        if not isinstance(arg_data, str):
                            raise AAZInvalidArgValueError(
                                f"Invalid format: Failed to construct resource id: argument '{arg_name}' is not string")

                        data_segments.append(arg_data)
                    else:
                        data_segments.append(segment)
                data = '/'.join(data_segments)
                # return data

            # verify resource id
            if len(data_segments) != len(self._segments):
                raise AAZInvalidArgValueError(
                    f"Invalid format: resource id should be in '{self._template}' format."
                )
            for data_segment, segment in zip(data_segments, self._segments):
                if not segment.startswith('{') and data_segment.lower() != segment.lower():
                    raise AAZInvalidArgValueError(
                        f"Invalid format: resource id should be in '{self._template}' format."
                    )

            return data

    def __init__(self, template=None, cross_tenants=True):
        """

        :param template: template property is used to verify a resource Id or construct resource Id.
        :param cross_tenants: if cross_tenants is True, the resource id will apply to cross-tenants scenarios.
        """
        self._template = None
        if template:
            self._template = self._Template(template)
        self._cross_tenants = cross_tenants

    def __call__(self, ctx, value):
        from azure.mgmt.core.tools import parse_resource_id

        assert isinstance(value, AAZSimpleValue)

        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        if self._template:
            data = self._template(ctx, data)

        parsed_id = parse_resource_id(data)
        subscription_id = parsed_id.get('subscription', None)
        if subscription_id and self._cross_tenants:
            # update subscription_id to support cross tenants
            ctx.update_aux_subscriptions(subscription_id)

        value._data = data
        return value


class AAZSubscriptionIdArgFormat(AAZBaseArgFormat):

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)

        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        # convert subscription name to id
        subscriptions_list = ctx.profile.load_cached_subscriptions()
        sub_id = None
        match_val = data.lower()
        for sub in subscriptions_list:
            # match by id
            if sub['id'].lower() == match_val:
                sub_id = sub['id']
                break
        if not sub_id:
            for sub in subscriptions_list:
                # match by name
                if sub['name'].lower() == match_val:
                    sub_id = sub['id']
                    break
        if sub_id:
            ctx.update_aux_subscriptions(sub_id)
            value._data = sub_id
        else:
            logger.warning("Subscription '%s' not recognized.", value._data)
        return value


class AAZFileArgFormat(AAZBaseArgFormat):

    def __call__(self, ctx, value):
        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)

        if not os.path.isfile(data):
            raise AAZInvalidArgValueError("File '{}' doesn't exist".format(data))

        data = self.read_file(data)
        value._data = data
        return value

    @abc.abstractmethod
    def read_file(self, file_path):
        raise NotImplementedError()


class AAZFileArgTextFormat(AAZFileArgFormat):

    def __init__(self, encoding=None):
        self._encoding = encoding  # use platform default encoding when it's None

    def read_file(self, file_path):
        with open(file_path, 'r', encoding=self._encoding) as f:
            data = f.read()
        return data


class AAZFileArgBase64EncodeFormat(AAZFileArgFormat):

    def read_file(self, file_path):
        with open(file_path, 'rb') as f:
            contents = f.read()
            base64_data = base64.b64encode(contents)
            try:
                data = base64_data.decode('utf-8')
            except UnicodeDecodeError:
                data = str(base64_data)
        return data


class AAZPaginationTokenArgFormat(AAZBaseArgFormat):
    def __call__(self, ctx, value):
        def validate_json(s):
            try:
                obj = json.loads(s)
            except json.JSONDecodeError:
                raise AAZInvalidArgValueError("Invalid JSON object.")

            if not isinstance(obj, dict):
                raise AAZInvalidArgValueError("Decoded object is not a dictionary.")

            try:
                _, _ = obj["next_link"], obj["offset"]
            except KeyError:
                raise AAZInvalidArgValueError("`next_link` or `offset` doesn't exist.")

        assert isinstance(value, AAZSimpleValue)
        data = value._data
        if data == AAZUndefined or data is None or value._is_patch:
            return value

        assert isinstance(data, str)
        try:
            decoded_bytes = base64.b64decode(data)
        except base64.binascii.Error:
            raise AAZInvalidArgValueError("Invalid Base64 string.")

        try:
            decoded_string = decoded_bytes.decode("utf-8")
        except UnicodeDecodeError:
            raise AAZInvalidArgValueError("Error decoding UTF-8.")

        validate_json(decoded_string)
        value._data = decoded_string

        return value



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "managedservices assignment delete",
    confirmation="Are you sure you want to perform this operation?",
)
class Delete(AAZCommand):
    """Delete a registration assignment.

    :example: Delete an assignment given its identifier under the default subscription scope.
        az managedservices assignment delete --assignment d3087cf0-e180-4cca-b147-54ae00c7b504

    :example: Delete an assignment given its fully qualified resource id
        az managedservices assignment delete --assignment /subscriptions/a62076fa-768a-403c-9d9d-6a9919aae441/providers/Microsoft.ManagedServices/registrationAssignments/0c3e9687-b461-4615-b6e4-74d54998d6e4
    """

    _aaz_info = {
        "version": "2019-06-01",
        "resources": [
            ["mgmt-plane", "/{scope}/providers/microsoft.managedservices/registrationassignments/{}", "2019-06-01"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, None)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.assignment = AAZStrArg(
            options=["--assignment"],
            help="The identifier (guid) or the fully qualified resource id of the registration assignment. When resource id is used, subscription id and resource group parameters are ignored.",
            required=True,
        )
        _args_schema.scope = AAZStrArg(
            options=["--scope"],
            help="Scope of the resource.",
            required=True,
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.RegistrationAssignmentsDelete(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    class RegistrationAssignmentsDelete(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [204]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_204,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/{scope}/providers/Microsoft.ManagedServices/registrationAssignments/{registrationAssignmentId}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "DELETE"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "registrationAssignmentId", self.ctx.args.assignment,
                    required=True,
                ),
                **self.serialize_url_param(
                    "scope", self.ctx.args.scope,
                    skip_quote=True,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2019-06-01",
                    required=True,
                ),
            }
            return parameters

        def on_200(self, session):
            pass

        def on_204(self, session):
            pass


class _DeleteHelper:
    """Helper class for Delete"""


__all__ = ["Delete"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long

def load_command_table(self, _):
    # with self.command_group('netappfiles account'):
    #     from .custom import AccountCreate, AccountUpdate
    #     self.command_table["netappfiles account create"] = AccountCreate(loader=self)
    #     self.command_table["netappfiles account update"] = AccountUpdate(loader=self)

    with self.command_group('netappfiles account ad'):
        from .custom import ActiveDirectoryAdd, ActiveDirectoryList, ActiveDirectoryUpdate
        self.command_table["netappfiles account ad add"] = ActiveDirectoryAdd(loader=self)
        self.command_table["netappfiles account ad update"] = ActiveDirectoryUpdate(loader=self)
        self.command_table["netappfiles account ad list"] = ActiveDirectoryList(loader=self)

    with self.command_group('netappfiles pool'):
        from .custom import PoolCreate, PoolUpdate
        self.command_table["netappfiles pool create"] = PoolCreate(loader=self)
        self.command_table["netappfiles pool update"] = PoolUpdate(loader=self)

    with self.command_group('netappfiles volume'):
        from .custom import VolumeCreate, VolumeUpdate, VolumeBreakFileLocks
        self.command_table["netappfiles volume create"] = VolumeCreate(loader=self)
        self.command_table["netappfiles volume update"] = VolumeUpdate(loader=self)
        self.command_table["netappfiles volume break-file-locks"] = VolumeBreakFileLocks(loader=self)

    with self.command_group('netappfiles volume-group'):
        from .custom import VolumeGroupCreate
        self.command_table["netappfiles volume-group create"] = VolumeGroupCreate(loader=self)

    with self.command_group('netappfiles volume export-policy'):
        from .custom import ExportPolicyList, ExportPolicyAdd, ExportPolicyRemove
        self.command_table['netappfiles volume export-policy list'] = ExportPolicyList(loader=self)
        self.command_table['netappfiles volume export-policy add'] = ExportPolicyAdd(loader=self)
        self.command_table['netappfiles volume export-policy remove'] = ExportPolicyRemove(loader=self)

    with self.command_group('netappfiles volume replication'):
        from .custom import ReplicationResume
        self.command_table['netappfiles volume replication resume'] = ReplicationResume(loader=self)

    with self.command_group('netappfiles', is_preview=False):
        from .custom import UpdateNetworkSiblingSet
        self.command_table["netappfiles update-network-sibling-set"] = UpdateNetworkSiblingSet(loader=self)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long, too-many-lines
from knack.arguments import CLIArgumentType

from azure.cli.core.commands.parameters import get_location_type, tags_type

name_arg_type = CLIArgumentType(options_list=('--name', '-n'), metavar='NAME',
                                help='The name of the identity resource.')


def load_arguments(self, _):

    with self.argument_context('identity') as c:
        c.argument('resource_name', arg_type=name_arg_type, id_part='name')

    with self.argument_context('identity create') as c:
        c.argument('location', get_location_type(self.cli_ctx), required=False)
        c.argument('tags', tags_type)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands import CliCommandType

from ._client_factory import _msi_user_identities_operations, _msi_operations_operations

from ._validators import process_msi_namespace


def load_command_table(self, _):

    identity_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.msi.operations#UserAssignedIdentitiesOperations.{}',
        client_factory=_msi_user_identities_operations
    )
    msi_operations_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.msi.operations#Operations.{}',
        client_factory=_msi_operations_operations
    )
    with self.command_group('identity', identity_sdk, client_factory=_msi_user_identities_operations) as g:
        g.custom_command('create', 'create_identity', validator=process_msi_namespace)
        g.show_command('show', 'get')
        g.command('delete', 'delete')
        g.custom_command('list', 'list_user_assigned_identities')
        g.custom_command('list-resources', 'list_identity_resources', min_api='2021-09-30-preview')

    with self.command_group('identity', msi_operations_sdk, client_factory=_msi_operations_operations) as g:
        g.command('list-operations', 'list')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command_group(
    "storage blob",
)
class __CMDGroup(AAZCommandGroup):
    """Manage object storage for unstructured data (blobs).

    Please specify one of the following authentication parameters for your commands: --auth- mode, --account-key, --connection-string, --sas-token. You also can use corresponding environment variables to store your authentication credentials, e.g. AZURE_STORAGE_KEY, AZURE_STORAGE_CONNECTION_STRING and AZURE_STORAGE_SAS_TOKEN.
    """
    pass


__all__ = ["__CMDGroup"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def list_user_assigned_identities(cmd, resource_group_name=None):
    from azure.cli.command_modules.identity._client_factory import _msi_client_factory
    client = _msi_client_factory(cmd.cli_ctx)
    if resource_group_name:
        return client.user_assigned_identities.list_by_resource_group(resource_group_name)
    return client.user_assigned_identities.list_by_subscription()


def create_identity(client, resource_group_name, resource_name, location, tags=None):
    parameters = {}
    parameters['location'] = location
    if tags is not None:
        parameters['tags'] = tags
    return client.create_or_update(resource_group_name=resource_group_name,
                                   resource_name=resource_name,
                                   parameters=parameters)


def list_identity_resources(cmd, resource_group_name, resource_name):
    from azure.cli.command_modules.identity._client_factory import _msi_list_resources_client
    client = _msi_list_resources_client(cmd.cli_ctx)
    return client.list_associated_resources(resource_group_name=resource_group_name,
                                            resource_name=resource_name)
